#ifndef _PWEVENT_H
#define _PWEVENT_H

/*
	ProWesS, PROGS Windowing System

	Definition of built in types for event handling

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 7 January 1995
	20.03.1996 reverted to using pixel coordinates, should be faster
*/

#include "io_h"
#ifndef _PWOBJECT_H
#include "PWObject_h"
#endif _PWOBJECT_H

typedef unsigned char Scale;		/* scale factor */
typedef int pix;					/* pixel values */

#define PW_EVENT_HIT		0x0001	/* call handler on hit event */
#define PW_EVENT_DO 		0x0002	/* call handler on do event */
#define PW_EVENT_TIMEOUT	0x0004	/* call handler on timeout event */
#define PW_EVENT_MOVE		0x0008	/* call handler on move event */
#define PW_EVENT_HITDRAG	0x0010	/* test for dragging on hit */
#define PW_EVENT_DODRAG 	0x0020	/* test for dragging on do */
						 /* 0x0040	/* unused */
#define PW_EVENT_DRAGEND	0x0080	/*+dragging has ended */
#define PW_EVENT_DRAGADJUST 0x0100	/* pointer adjusted to stay in region */
#define PW_EVENT_CATCH		0x0200	/*+catch untrapped keypress */
#define PW_EVENT_CATCHSTART 0x0400	/*+object is selected to start catching */
#define PW_EVENT_CATCHEND	0x0800	/*+object is deselected from catching */
typedef short PWWait;				/* type for these events (+always active) */


struct _PWRegion {
	PWSystem *system;				/* global data structure for window */
	PWObject parentObject;			/* parent (owner when creating file) */
	PWObject owner; 				/* owner/creator of this region */
	struct _PWRegion *parent;		/* parent region */
	struct _PWRegion *next; 		/* next region on this level */
	struct _PWRegion *prev; 		/* previous region on this level */
	struct _PWRegion *children; 	/* list on next level of nesting */
	short level;					/* nesting level (==parent->level+1)
									/* parameters for region handler */
	PWWait wait_event;				/* events to wait for */
	TimeOut timeout;				/* timeout for events */
									/* total area covered by region */
	pt tot_xpos, tot_ypos;			/* origin of region in parent */
	pt tot_xsiz, tot_ysiz;			/* size of region */
									/* hit area, only this area is handled by */
									/* the object type handle routine */
	pix hit_off_lft, hit_off_rgt;	/* margins of hit area in total area */
	pix hit_off_top, hit_off_bot;	/* margins of hit area in total area */
	pix hit_xorg, hit_yorg; 		/* top left coordinate of area on screen */
	pix hit_xend, hit_yend; 		/* bottom right coordinate */
									/* scale parameters for object */
									/* the size of the margins around the hit */
									/* area stays the same */
	Scale scale;					/* scale factor for size (primary dir.) */
	char windowfit; 				/* secondary direction */
	char composite; 				/* this is a composite object */
	char dragging;					/* are we dragging in this region ?? */
	};

struct _PWKeyPress {
	PWSystem *system;				/* global data structure for window */
	PWObject parentObject;			/* parent (owner when creating file */
	PWObject owner; 				/* owner/creator of this keypress */
	struct _PWKeyPress *next;		/* next keypress in list of keyspresses */
	char key;						/* key which should be pressed */
	char alternative;				/* alternative for this action */
	};

typedef struct _PWRegion PWRegion;
typedef struct _PWKeyPress PWKeyPress;

#endif _PWEVENT_H
