#ifndef _PWHANDLER_H
#define _PWHANDLER_H

/*
	ProWesS, PROGS Windowing System

	Definition of functions for writing object type handlers

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 21 January 1995
*/

#ifndef _PROWESS_H
#include "ProWesS_h"
#endif
#ifndef _PWOBJECT_H
#include "PWObject_h"
#endif
#ifndef _PWSYSTEM_H
#include "PWSystem_h"
#endif
#ifndef _PWTYPEDEF_H
#include "PWTypeDef_h"
#endif


Error PWDetermineSize(PWSystem *system);
Error PWRemoveIndicated(PWSystem *system);
Error PWHandleKeyPress(PWSystem *system);
Error PWHandleRegion(PWSystem *system, PWRegion *region);
Error PWRedraw(PWSystem *system);
void PWDeterminePosition(PWSystem *system, int xborder, int yborder);
Error PWWindowFit(PWSystem *system, pt xinc, pt yinc);
Error PWSetWindow(PWSystem *system);
Error PWSetSprites(PWSystem *system);
Error PWSleep(PWSystem *system);

#define PW_CORNER_TOPLEFT	0
#define PW_CORNER_TOPRIGHT	2
#define PW_CORNER_BOTLEFT	4
#define PW_CORNER_BOTRIGHT	6
Error PWWindowMove(PWSystem *system,short move);
Error PWWindowScale(PWSystem *system,short scale);
Error PWScrollScreen(PWSystem *system, int x, int y, char move);
Error PWScrollTouch(PWSystem *system);
Error SoftwareEvent(PWSystem *system);

void PWpixel2point(pt *xfactor, pt *yfactor);
void PWpoint2pixel(pt *xfactor, pt *yfactor);
void PWpixel2virtual(pt *xfactor, pt *yfactor);
void PWvirtual2pixel(pt *xfactor, pt *yfactor);

Error PWGetType(PWType type, PWTypeDef **rettype);
Error PWClose(PWSystem *system);

int *PWSkipTag(int *taglist);
Error PWAllocate(PWSystem *system, Size size, void **ret);
Error PWRelease(PWObject object, void *base);

int PF2xpix(pt x);
int PF2ypix(pt x);
pt xpix2PF(short x);
pt ypix2PF(short x);
pt PFxround(pt x);
pt PFyround(pt x);

typedef struct _Defines {
	char *name; 					/* name of constant, no prefix */
	Error (*Handle)(struct _Defines *possible, char *value);
	void (*Post)(void *data);		/* post processing routine */
	void *data; 					/* place where data should be put */
	char *set;						/* set to TRUE when data set (if !NULL) */
} Defines;

#define PW_DEFINE_COLOUR		((Error (*)(Defines*,char*))1)
#define PW_DEFINE_BOOLEAN		((Error (*)(Defines*,char*))2)
#define PW_DEFINE_CHARACTER 	((Error (*)(Defines*,char*))3)
#define PW_DEFINE_SHORT 		((Error (*)(Defines*,char*))4)
#define PW_DEFINE_INT			((Error (*)(Defines*,char*))5)
#define PW_DEFINE_PT			((Error (*)(Defines*,char*))6)
#define PW_DEFINE_FONT			((Error (*)(Defines*,char*))7)

Error PWHandleDefine(char *prefix, Defines *possible, char *name, char *value);

#endif _PWHANDLER_H
