#ifndef _PWSYSTEM_H
#define _PWSYSTEM_H

/*
	ProWesS, PROGS Windowing System

	Definition of the internal system

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 7 January 1995
	20.03.1996 reverted to using pixl coordinates, should be faster
*/

#include "mem_h"
#include "win_h"
#include "job_h"

#ifndef _PWEVENT_H
#include "PWEvent_h"
#endif

/* Each type can have a general state, to keep information about it */
/* Normally the base type is followed by extra type specific information */
/* The PWTypeState information is kept in a list with sentinel (type=NULL) */

#define PWBaseTypeState 									\
	PWType type;					/* type identifier */	\
	struct _PWDummyTypeState *next	/* next item */

typedef struct _PWDummyTypeState {
	PWBaseTypeState;				/* common start for all typestates */
	} PWDummyTypeState;

/* System data structure */

struct _PWSystem {
	struct _PWSystem *next; 		/* next PWSystem for this job */

	PWRegion *regions;				/* tree of info regions in window */
	PWKeyPress *keypresses; 		/* list of keypresses to scan */

	Window window;					/* Window identifier */
									/* data for WINPointerRead */
	PointerInfo ptr_info;			/* the current pointer record */
	short ptr_xpos, ptr_ypos;		/* current pointer position */

	void *global;					/* user fillable data - global structure */
	PWDummyTypeState *state;		/* state information for types used */
	PWObject catchkeys; 			/* object which has to catch keypresses */
	char catch; 					/* key for which catching is needed */
	char activated; 				/* TRUE when system activated */

	unsigned long status;			/* system status (values WSTATUS_xxx) */
									/* information about the window/outline */
	pix min_xsiz, min_ysiz; 		/* minimum window size */
	pix max_xsiz, max_ysiz; 		/* maximum window size */
	pix curr_xsiz, curr_ysiz;		/* current window size */
	pix xorg, yorg; 				/* origin of area in window */
	pix win_xpos, win_ypos; 		/* position of window */

	Gstate gstate;					/* gstate when PROforma is used (or NULL) */

	PWRegion *sleep;				/* sleep region */
	PWRegion *active;				/* currently active region */

	Priority priority;				/* previous priority for SSTATUS_BUTFRAME */

	int Xdragadjust, Ydragadjust; /* adjustment of pointer while dragging */

	SubWindowList subwindows;		/* list of subwindows */
	SubWindow *subwinlist[9];		/* array with pointers to subwins */
	SubWindow subwin[8];			/* the definition of the subwindows */

	Heap heap;						/* heap with/for this system */
	Error (*ActionInit)(PWObject object);
	Error (*EventLevel1)(PWObject object);
	Error (*EventLevel2)(PWObject object);
	Error (*EventLevel3)(PWObject object);
	Error (*EventLevel4)(PWObject object);
	Error (*EventLevel5)(PWObject object);
	Error (*EventLevel6)(PWObject object);
	Error (*EventLevel7)(PWObject object);
	Error (*EventLevel8)(PWObject object);
	};

#define SSTATUS_REDRAW		0x00000001	/* set when part of window needs redraw */
#define SSTATUS_BREAKDOWN	0x00000002	/* set when activation should be ended */
#define SSTATUS_WINCHANGE	0x00000004	/* set when window build has changed */
#define SSTATUS_SLEEP		0x00000008	/* set when window should be put asleep */
#define SSTATUS_ALLDRAW 	0x00000010	/* set when all of window needs redraw */
#define SSTATUS_REMOVE		0x00000020	/* set when an object has to be removed */
#define SSTATUS_PAGESHOW	0x00000040	/* set when PFPageShow should be done */

#define SSTATUS_NOXSCALE	0x00000100	/* no horizontal scaling */
#define SSTATUS_NOYSCALE	0x00000200	/* no vertical scaling */

#define SSTATUS_NOXBORDER	0x00010000	/* no scaleborder - horizontal */
#define SSTATUS_NOYBORDER	0x00020000	/* no scaleborder - vertical */
#define SSTATUS_XSCROLLABLE 0x00040000	/* window is larger than screen - x */
#define SSTATUS_YSCROLLABLE 0x00080000	/* window is larger than screen - y */
#define SSTATUS_BUTFRAME	0x00100000	/* put window in button frame */

#endif _PWSYSTEM_H
