#ifndef _ProWesS_H
#define _ProWesS_H

/*
	ProWesS, PROGS Windowing System

	Standard (public) definitions.

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 7 January 1995
*/

#include "sms_h"
#include "err_h"

#ifndef _PROFORMA_H
#include "PROforma_h"
#endif

/* general types which are visible to the ProWesS user */

typedef struct _PWDummyObject *PWObject;
typedef unsigned long PWType;
typedef struct _PWTypeDef PWTypeDef;

/* prototypes for public routines */
Error PWCreate(PWObject owner, PWObject *ret, PWType type, ...);
Error PWChange(PWObject object, ...);
Error PWDoChange(PWObject object, int *taglist);
Error PWQuery(PWObject object, int what, void *ret);
Error PWActivate(PWObject object);
Error PWAddType(PWTypeDef *newtype);
Error PWRemove(PWObject object);

/* access to ProWesS thing for dynamic configuration */
#define PW_THING_NAME	"ProWesS"
#define PW_THING_CNFG	0x434e4647	/* "CNFG" */

/* TAGS, these are used to identify which parameters will be changed */
/* The tags are 32 bit numbers. However, the top 20 bits are reserved to */
/* give extra information. There are two sets of tags, general tags, which */
/* can apply to any object/type, and (type) specific tags, which only apply
/* to certain objects/types. There should be no overlap between */
/* the general and specific tags. However specific tags for different types */
/* may overlap (though this may introduce unwanted synonyms). Therefore, it */
/* is adviced that anybody who distributes new types acquires a mask for the */
/* tags, to indicate the software producer. */
/* The tag contains 3 fields : */
/* tag identifier, 8 bits, identifies length of parameters */
/* manufacturer, 12 bits, identifies instances which determined the tag number */
/*						  value : 0 for general, 1 for standard */
/* tag number, 12 bits, identifies the use of the tag, has to be non-zero */

#define PW_TAGMASK_LIST 		0x80000000	/* parameter is a NULL terminated list */
#define PW_TAGMASK_TAGLIST		0x40000000	/* parameter is a NULL terminated list of tags */
#define PW_TAGMASK_MULBIT		0x10000000	/* bit set when skip specific */
#define PW_TAGMASK_MULTIPLE(x)	((0x10+x)<<24)/* parameter is "x" values */
#define PW_TAGINFO_SKIP(x)		(((x)>>24)&0xf) /* extract skip from tag */
#define PW_TAGMASK_SKIP 		0xff000000	/* skip this tag and parameter */
#define PW_TAGMASK_PRODUCER 	0x00fff000	/* manufacturer type */

/* general tags */

#define PW_DUMMY				(0x000 +PW_TAGMASK_MULTIPLE(0))
#define PW_POSITION_RIGHT_OF	(0x001 +PW_TAGMASK_MULTIPLE(1))
#define PW_POSITION_LEFT_OF 	(0x002 +PW_TAGMASK_MULTIPLE(1))
#define PW_POSITION_ABOVE		(0x003 +PW_TAGMASK_MULTIPLE(1))
#define PW_POSITION_BELOW		(0x004 +PW_TAGMASK_MULTIPLE(1))
#define PW_POSITION_NEXT_ROW	(0x005 +PW_TAGMASK_MULTIPLE(0))
#define PW_POSITION_NEXT_COLUMN (0x006 +PW_TAGMASK_MULTIPLE(0))
#define PW_OBJECT_AUXILIARY 	(0x007 +PW_TAGMASK_MULTIPLE(1))
#define PW_GLOBAL_AUXILIARY 	(0x008 +PW_TAGMASK_MULTIPLE(1))
#define PW_SYSTEM_BREAKDOWN 	(0x009 +PW_TAGMASK_MULTIPLE(0))
#define PW_POINTER				(0x00a +PW_TAGMASK_MULTIPLE(1))
#define PW_WINDOWFIT			(0x00b +PW_TAGMASK_MULTIPLE(1))
#define PW_SCALE_FACTOR 		(0x00c +PW_TAGMASK_MULTIPLE(1))
#define PW_WINDOW_SIZE			(0x00d +PW_TAGMASK_MULTIPLE(2))
#define PW_WINDOW_XSIZE 		(0x00e +PW_TAGMASK_MULTIPLE(1))
#define PW_WINDOW_YSIZE 		(0x00f +PW_TAGMASK_MULTIPLE(1))
#define PW_WINDOW_ORIGIN		(0x010 +PW_TAGMASK_MULTIPLE(2))
#define PW_WINDOW_XORIGIN		(0x011 +PW_TAGMASK_MULTIPLE(1))
#define PW_WINDOW_YORIGIN		(0x012 +PW_TAGMASK_MULTIPLE(1))
#define PW_WINDOW_SIZE_PIX		(0x013 +PW_TAGMASK_MULTIPLE(2))
#define PW_WINDOW_XSIZE_PIX 	(0x014 +PW_TAGMASK_MULTIPLE(1))
#define PW_WINDOW_YSIZE_PIX 	(0x015 +PW_TAGMASK_MULTIPLE(1))
#define PW_WINDOW_ORIGIN_PIX	(0x016 +PW_TAGMASK_MULTIPLE(2))
#define PW_WINDOW_XORIGIN_PIX	(0x017 +PW_TAGMASK_MULTIPLE(1))
#define PW_WINDOW_YORIGIN_PIX	(0x018 +PW_TAGMASK_MULTIPLE(1))
#define PW_KEYPRESS 			(0x019 +PW_TAGMASK_MULTIPLE(1))
#define PW_SLEEP_OBJECT 		(0x01a +PW_TAGMASK_MULTIPLE(0))
#define PW_SYSTEM_SLEEP 		(0x01b +PW_TAGMASK_MULTIPLE(0))
#define PW_WINDOW_BUTTON		(0x01c +PW_TAGMASK_MULTIPLE(0))
#define PW_WINDOW_NOSCALE		(0x01d +PW_TAGMASK_MULTIPLE(0))
#define PW_WINDOW_NOXSCALE		(0x01e +PW_TAGMASK_MULTIPLE(0))
#define PW_WINDOW_NOYSCALE		(0x01f +PW_TAGMASK_MULTIPLE(0))
#define PW_WINDOW_ORIGIN_RESET	(0x020 +PW_TAGMASK_MULTIPLE(0))
#define PW_CURSOR_SEPARATE		(0x020 +PW_TAGMASK_MULTIPLE(1))
#define PW_CATCH_OBJECT 		(0x021 +PW_TAGMASK_MULTIPLE(1))
#define PW_SYSTEM_GSTATE		(0x022 +PW_TAGMASK_MULTIPLE(0))
#define PW_SYSTEM_ACTION_INIT	(0x023 +PW_TAGMASK_MULTIPLE(1))
#define PW_SYSTEM_ACTION_EVENT1 (0x024 +PW_TAGMASK_MULTIPLE(1))
#define PW_SYSTEM_ACTION_EVENT2 (0x025 +PW_TAGMASK_MULTIPLE(1))
#define PW_SYSTEM_ACTION_EVENT3 (0x026 +PW_TAGMASK_MULTIPLE(1))
#define PW_SYSTEM_ACTION_EVENT4 (0x027 +PW_TAGMASK_MULTIPLE(1))
#define PW_SYSTEM_ACTION_EVENT5 (0x028 +PW_TAGMASK_MULTIPLE(1))
#define PW_SYSTEM_ACTION_EVENT6 (0x029 +PW_TAGMASK_MULTIPLE(1))
#define PW_SYSTEM_ACTION_EVENT7 (0x02a +PW_TAGMASK_MULTIPLE(1))
#define PW_SYSTEM_ACTION_EVENT8 (0x02b +PW_TAGMASK_MULTIPLE(1))

/* tags to query for default values which are stored in ProWesS system */
#define PW_DEFAULT_BACKGROUND	(0x100 +PW_TAGMASK_MULTIPLE(1))
#define PW_DEFAULT_FOREGROUND	(0x101 +PW_TAGMASK_MULTIPLE(1))
#define PW_DEFAULT_MIDDLEGROUND (0x102 +PW_TAGMASK_MULTIPLE(1))
#define PW_DEFAULT_FONT 		(0x103 +PW_TAGMASK_MULTIPLE(1))
#define PW_DEFAULT_FONTSIZE 	(0x104 +PW_TAGMASK_MULTIPLE(1))

#include "PWStdTags_h"

#endif _ProWesS_H
