/*
	ProWesS, PROGS Windowing System

	direct edit line of text type for ProWesS, similar to "EDLI", but no need
	to indicate it to start typing, just go ahead "DEDL"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 11 September 1995, based on edline
	26.10.1995 made sure escape codes are possible when no cursor is visible
	22.03.1996 adjusted for internal use of pixels
	11.04.1996 use PWHandleDefine && use ProWesS defaults
	11.05.1996 PW_EDLINE_ACTIVATE added, selectes this object as catch object
	08.06.1996 PW_EDLINE_ACTION_TEST added, including three default actions
	02.07.1996 PW_EDLINE_ACTION_TEST routine now also called before first key
*/

#include "mem_h"
#include "str_h"
#include "ctype_h"
#include "io_h"

#include "PWHandler_h"

#define TYPE PW_TYPE_DIRECT_EDLINE

#define catch(X) if (err=(X)) return err; else

/* global statusses */

#define p100 short2pt(100)
ColourRGB ink	= { 0,    0,    0    }; char Kink  =FALSE;
ColourRGB paper = { p100, p100, p100 }; char Kpaper=FALSE;
char loose_font[PF_MAXFONTNAME]=""; /* name of PROforma font to use */
pt loose_fontsize=0;				/* fontsize to use */
pt loose_baseline=0;				/* font baseline position (at fontsize) */
pt loose_height=0;					/* height of font (at fontsize) */
short bufflen=256;					/* default maximum size of strings */
pt xsiz=short2pt(120);				/* default minimum size of edline object */
char recalculate=TRUE;				/* set when baseline and height should be determined */

/* loose item object definition */

typedef struct _Object {
	PWBaseObject;
	int bufflen;			/* length of buffer */
	pix xsiz;				/* minimum width of item */
	PWObject keypress;		/* keypress attached to this object */
	Error (*DoAfter)(PWObject object);	/* call after pressing <enter> */
	Error (*ActionTest)(PWObject object, EdlineInfo *info);
	struct _Object *next, *prev, *up, *down, *after;
	int cursor; 			/* cursor position, before string[cursor] */
	pt curpos;				/* cursor position, in window */
	pt offset;				/* x pageorigin where string was displayed */
	int length; 			/* length of string */
	char string[2]; 		/* string to edit - length==bufflen */
	} Object;

/* prototypes */
Error Handle(PWObject object, PWWait event);

/* object handling routines */

void Calculate(Gstate gstate, PWObject object)
{
	Bbox bbox;
	recalculate=FALSE;
	if (!Kink) PWQuery(object,PW_DEFAULT_FOREGROUND,&ink);
	if (!Kpaper) PWQuery(object,PW_DEFAULT_BACKGROUND,&paper);
	if (!loose_fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&loose_fontsize);
	if (!*loose_font) PWQuery(object,PW_DEFAULT_FONT,loose_font);
	PFFontLoad(gstate,loose_font);
	PFFontScale(gstate,loose_fontsize);
	PFFontBbox(gstate,&bbox);
	loose_baseline=bbox.yorg;
	loose_height=bbox.ysiz;
}

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Error err;
	short buffer=bufflen;

	LinkPROforma();

	/* see if there is a PW_EDLINE_MAXLENGTH tag */
	while (*taglist)
	{
		if (*taglist==PW_EDLINE_MAXLENGTH)
		{
			buffer=(*(taglist+1));
			*taglist&=PW_TAGMASK_SKIP;
		}
		taglist=PWSkipTag(taglist);
	}

	/* allocate specific object - which should be cleared (automatic) */
	catch( PWAllocate(event.region->system,sizeof(Object)+buffer,ret) );

	/* make sure the item is available */
	(*(Object **)ret)->bufflen=buffer;
	(*(Object **)ret)->xsiz=PF2xpix(xsiz);
	(*(Object **)ret)->cursor=-1;
	
	return err;
}

Error Remove(PWObject object)
{
	Object *obj=(Object *)object;
	if (obj->keypress) PWRemove(obj->keypress);
	return PWRelease((PWObject)obj,(PWObject)obj);
}

Error select(PWObject object)
{
	PWObject owner;
	PWQuery(object,PW_OBJECT_AUXILIARY,&owner);
	return PWChange(object,PW_CATCH_OBJECT,owner,NULL);
}

static Error natural(PWObject object, EdlineInfo *info)
{
	if (isprint(info->key) && !isdigit(info->key))
		return PW_EDLINE_TEST_NOACCEPT;
	else
		return PW_EDLINE_TEST_ACCEPT;
}

static Error integer(PWObject object, EdlineInfo *info)
{
	if (info->key=='+' || info->key=='-')
	{
		if (info->string[0]=='+' || info->string[0]=='-')
		{
			info->string[0]=info->key;
			return PW_EDLINE_TEST_CHANGE;
		}
		if (info->length+1<info->maxlength)
		{
			MEMCopy(info->maxlength-1,info->string,info->string+1);
			info->string[0]=info->key;
			info->cursor++;
			info->length++;
			return PW_EDLINE_TEST_CHANGE;
		}
		return PW_EDLINE_TEST_NOACCEPT;
	}
	if (isprint(info->key) && !isdigit(info->key))
		return PW_EDLINE_TEST_NOACCEPT;
	else
		return PW_EDLINE_TEST_ACCEPT;
}

static Error number(PWObject object, EdlineInfo *info)
{
	if (info->key=='+' || info->key=='-')
	{
		if (info->string[0]=='+' || info->string[0]=='-')
		{
			info->string[0]=info->key;
			return PW_EDLINE_TEST_CHANGE;
		}
		if (info->length+1<info->maxlength)
		{
			MEMCopy(info->maxlength-1,info->string,info->string+1);
			info->string[0]=info->key;
			info->cursor++;
			info->length++;
			return PW_EDLINE_TEST_CHANGE;
		}
		return PW_EDLINE_TEST_NOACCEPT;
	}
	if (info->key=='.')
	{
		/* only accept if no decimal point yet */
		char *str;
		for (str=info->string ; *str ; str++)
			if (*str=='.') return PW_EDLINE_TEST_NOACCEPT;
		return PW_EDLINE_TEST_ACCEPT;
	}
	if (isprint(info->key) && !isdigit(info->key))
		return PW_EDLINE_TEST_NOACCEPT;
	else
		return PW_EDLINE_TEST_ACCEPT;
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	Gstate gstate=system->gstate;
	char unrecognised_tag=FALSE;

	if (!(obj->status&OSTATUS_INITIALISED))
	{
		obj->status|=OSTATUS_INITIALISED;
		PWChange((PWObject)obj,PW_CURSOR_SEPARATE,TRUE,NULL);
		if (system->catchkeys==NULL)
			PWChange((PWObject)obj,PW_CATCH_OBJECT,obj,NULL);
	}

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_EDLINE_SET:
			STRCopySafe((char *)(*(taglist+1)),obj->string,obj->bufflen);
			Handle((PWObject)obj,PW_EVENT_CATCHSTART);
			break;
		case PW_EDLINE_GET:
			/* two parameters : buffer length and buffer base */
			STRCopySafe(obj->string,(char *)(*(taglist+2)),(*(taglist+1)));
			break;
		case PW_EDLINE_WIDTH:
			obj->xsiz=PF2xpix(*(taglist+1));
			system->status|=SSTATUS_WINCHANGE;
			break;
		case PW_EDLINE_WIDTH_PIX:
			obj->xsiz=(*(taglist+1));
			system->status|=SSTATUS_WINCHANGE;
			break;
		case PW_EDLINE_WIDTH_FS:
			obj->xsiz=PF2xpix(fixmul((pt)(*(taglist+1)),loose_fontsize));
			system->status|=SSTATUS_WINCHANGE;
			break;
		case PW_EDLINE_KEYPRESS:
			/* attach keypress */
			if (!obj->keypress)
				catch( PWCreate(object,&obj->keypress,PW_TYPE_KEYPRESS,
								PW_KEYPRESS_ACTION, select,
								PW_OBJECT_AUXILIARY, object,
								NULL) );
			catch( PWChange(obj->keypress,PW_KEYPRESS,(*(taglist+1)),NULL) );
			break;
		case PW_EDLINE_ACTION_AFTER:
			obj->DoAfter=(Error (*)(PWObject))(*(taglist+1));
			break;
		case PW_EDLINE_ACTION_TEST:
			if ((*(taglist+1))<0)
			{
				switch (*(taglist+1))
				{
				case PW_EDLINE_ACTION_TEST_NATURAL:
					obj->ActionTest=natural;
					break;
				case PW_EDLINE_ACTION_TEST_INTEGER:
					obj->ActionTest=integer;
					break;
				case PW_EDLINE_ACTION_TEST_FLOAT:
					obj->ActionTest=number;
					break;
				}
			}
			else
				obj->ActionTest=(Error (*)(PWObject, EdlineInfo *))(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_AFTER:
			obj->after=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_NEXT:
			obj->next=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_PREV:
			obj->prev=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_UP:
			obj->up=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_DOWN:
			obj->down=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_ACTIVATE:
			PWChange((PWObject)obj,PW_CATCH_OBJECT,obj,NULL);
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Error Draw(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *reg=obj->event.region;
	Gstate gstate=reg->system->gstate;
	pix hit_xsiz=reg->hit_xend-reg->hit_xorg;
	pix hit_ysiz=reg->hit_yend-reg->hit_yorg;
	pt ypos=(xpix2PF(hit_ysiz)-loose_height)/2;
	pt dist, min_cur, max_cur;

	if (!gstate) return ERR_OK;
	if (recalculate) Calculate(gstate,object);

	PFPWindowSubSet(gstate,hit_xsiz,hit_ysiz,reg->hit_xorg,reg->hit_yorg);
	PFPaperColourRGB(gstate,&paper);
	PFColourRGB(gstate,&ink);
	PFPageClear(gstate);
	PFFontLoad(gstate,loose_font);
	PFFontScale(gstate,loose_fontsize);

	if (obj->cursor>=0) PFPrintEscape(gstate,TRUE);

	min_cur=min(loose_fontsize,xpix2PF(hit_xsiz/8));
	max_cur=xpix2PF(hit_xsiz)-min_cur;

	/* make sure cursor is visible in window */
	if ( obj->cursor>=0 &&
		 ( (dist=obj->curpos-max_cur-obj->offset)>0 ||
		   (dist=obj->curpos-min_cur-obj->offset)<0 ) )
	{
		obj->offset+=dist;
		if (obj->offset<0)
			obj->offset=0;
	}
	PFPageOriginSet(gstate,obj->offset,0);
	PFMoveTo(gstate,0,ypos+loose_baseline);
	PFStringShow(gstate,obj->string);
	if (obj->cursor>=0)
	{
		PFPathStroked(gstate);
		PFLineWidth(gstate,0);
		PFMoveTo(gstate,obj->curpos,ypos);
		PFLineR(gstate,0,loose_height);
		PFPathDraw(gstate);
	}

	PFWindowSubRestore(gstate);
	PFPrintEscape(gstate,FALSE);

	return ERR_OK;
}

Error Handle(PWObject object, PWWait event)
{
	PWSystem *system=object->event.region->system;
	Gstate gstate=system->gstate;
	Object *obj=(Object *)object;

	if (!gstate) return ERR_OK;

	PFPrintEscape(gstate,TRUE);

	switch (event)
	{
	case PW_EVENT_CATCHSTART:
		if (recalculate) Calculate(gstate,(PWObject)obj);
		PFFontLoad(gstate,loose_font);
		PFFontScale(gstate,loose_fontsize);
		PFStringWidth(gstate,obj->string,&obj->curpos);
		obj->length=strlen(obj->string);
		obj->cursor=obj->length;
		system->catch=0;
		goto catch;
	case PW_EVENT_CATCHEND:
		obj->offset=0;
		obj->cursor=-1;
		goto redraw;
	case PW_EVENT_CATCH:
catch:	{
			char c=system->catch;
			char w[2]={0,0}; pt cw;
			Object *nextobj=NULL;

			/* handle the keypress, allow callback mechanism to check */
			/* whether the keypress is valid */
			/* The callback routine can have three possible results */
			/* ACCEPT keypress is valid, use normal processing */
			/* NO ACCEPT do not accept */
			/* CHANGE do not use normal processing, the string and/or */
			/*		cursor position are changed by the callback routine */
			if (obj->ActionTest)
			{
				int i;
				Error err;
				EdlineInfo info;

				info.maxlength=obj->bufflen;
				info.string=obj->string;
				info.length=obj->length;
				info.cursor=obj->cursor;
				info.key=c;
				err=obj->ActionTest((PWObject)obj,&info);
				switch(err)
				{
				case PW_EDLINE_TEST_NOACCEPT:
					return ERR_OK;
				case PW_EDLINE_TEST_ACCEPT:
					c=info.key;
					break;
				case PW_EDLINE_TEST_CHANGE:
					/* string and/or cursor position have changed */
					/* adjust internal structures */
					/* determine new value of currpos */
					obj->cursor=info.cursor;
					obj->length=info.length;
					obj->curpos=0;
					for (i=0 ; i<info.cursor ; i++)
					{
						w[0]=obj->string[i];
						PFStringWidth(gstate,w,&cw);
						obj->curpos+=cw;
					}
					goto redraw;
				}
			}

			/* put it in the string */
			if (isprint(c) && obj->length<obj->bufflen-1)
			{
				/* put it in the string */
				char *pos=&obj->string[obj->cursor];
				char *end=&obj->string[++obj->length];
				while (end>=pos) { *(end+1)=*end; end--; }
				*pos=c;
				obj->cursor++;
				w[0]=c;
				PFStringWidth(gstate,w,&cw);
				obj->curpos+=cw;
			}
			else switch(c)
			{
			case '\n':
				nextobj=obj->after;
				if (obj->DoAfter)
					obj->DoAfter((PWObject)obj);
				break;
			case '\t':
				nextobj=obj->next;
				break;
			case (char)0xfd:
				nextobj=obj->prev;
				break;
			case (char)0xd0:
				nextobj=obj->up;
				break;
			case (char)0xd8:
				nextobj=obj->down;
				break;
			case (char)0xc0 : /* left */
				if (obj->cursor)
				{
					obj->cursor--;
					w[0]=obj->string[obj->cursor];
					PFStringWidth(gstate,w,&cw);
					obj->curpos-=cw;
				}
				break;
			case (char)0xc1 : /* alt left - cursor to start of line */
				obj->cursor=0;
				obj->curpos=0;
				break;
			case (char)0xc2 : /* ctrl left - backspace */
				if (obj->cursor)
				{
					char *pos=&obj->string[obj->cursor];
					w[0]=*(pos-1);
					obj->cursor--;
					PFStringWidth(gstate,w,&cw);
					obj->curpos-=cw;
					while (*pos) { *(pos-1)=*pos; pos++; }
					*(pos-1)='\0';
					obj->length--;
				}
				break;
			case (char)0xc3 : /* ctrl alt left - delete start of line */
				{
					char *pos=obj->string;
					int i=obj->length-obj->cursor;
					for(;i;i--) *pos++=*(pos+obj->cursor);
					*pos=0;
					obj->length-=obj->cursor;
					obj->cursor=0;
					obj->curpos=0;
				}
				break;
			case (char)0xc4 : /* shift left */
				{
					int gotspace=obj->string[obj->cursor-1]==' ';
					while ( obj->cursor &&
							( (obj->string[obj->cursor-1]==' ' &&  gotspace) ||
							  (obj->string[obj->cursor-1]!=' ' && !gotspace) ) )
					{
						obj->cursor--;
						w[0]=obj->string[obj->cursor];
						PFStringWidth(gstate,w,&cw);
						obj->curpos-=cw;
						if (gotspace) gotspace=obj->string[obj->cursor-1]==' ';
					}
				}
				break;
			case (char)0xc6 : /* shift ctrl left */
				{
					int oldcursor=obj->cursor;
					int gotspace=obj->string[obj->cursor-1]==' ';
					while ( obj->cursor &&
							( (obj->string[obj->cursor-1]==' ' &&  gotspace) ||
							  (obj->string[obj->cursor-1]!=' ' && !gotspace) ) )
					{
						obj->cursor--;
						w[0]=obj->string[obj->cursor];
						PFStringWidth(gstate,w,&cw);
						obj->curpos-=cw;
						if (gotspace) gotspace=obj->string[obj->cursor-1]==' ';
					}
					if (obj->cursor!=oldcursor)
					{
						char *pos=&obj->string[obj->cursor];
						oldcursor-=obj->cursor;
						while (*pos++=*(pos+oldcursor));
						obj->length-=oldcursor;
					}
				}
				break;
			case (char)0xc8 : /* right */
				if (obj->cursor<obj->length)
				{
					w[0]=obj->string[obj->cursor];
					PFStringWidth(gstate,w,&cw);
					obj->cursor++;
					obj->curpos+=cw;
				}
				break;
			case (char)0xc9 : /* alt right */
				obj->cursor=obj->length;
				PFStringWidth(gstate,obj->string,&obj->curpos);
				break;
			case (char)0xca : /* ctrl right */
				{
					char *pos=&obj->string[obj->cursor];
					while (*pos++=*(pos+1));
					obj->length--;
				}
				break;
			case (char)0xcb : /* ctrl alt right */
				obj->string[obj->cursor]=0;
				obj->length=strlen(obj->string);
				break;
			case (char)0xcc : /* shift right */
				{
					int gotspace=obj->string[obj->cursor]==' ';
					while ( obj->cursor<obj->length &&
							( (obj->string[obj->cursor]==' ' &&  gotspace) ||
							  (obj->string[obj->cursor]!=' ' && !gotspace) ) )
					{
						w[0]=obj->string[obj->cursor];
						PFStringWidth(gstate,w,&cw);
						obj->cursor++;
						obj->curpos+=cw;
						if (!gotspace) gotspace=obj->string[obj->cursor]==' ';
					}
				}
				break;
			case (char)0xce : /* shift ctrl right */
				{
					int gotspace=obj->string[obj->cursor]==' ';
					int endpos=obj->cursor;
					while ( endpos<obj->length &&
							( (obj->string[endpos]==' ' &&	gotspace) ||
							  (obj->string[endpos]!=' ' && !gotspace) ) )
					{
						endpos++;
						if (!gotspace) gotspace=obj->string[endpos]==' ';
					}
					if (obj->cursor!=endpos)
					{
						char *pos=&obj->string[obj->cursor];
						endpos-=obj->cursor;
						while (*pos++=*(pos+endpos));
						obj->length-=endpos;
					}
				}
				break;
			case (char)0xda : /* ctrl down */
				obj->string[0]='\0';
				obj->curpos=obj->length=obj->cursor=0;
				break;
			}
			if (nextobj)
				PWChange((PWObject)obj,PW_CATCH_OBJECT,nextobj,NULL);
		}
redraw: if (system->activated)
		{
			Draw((PWObject)obj);
			PFPageShow(gstate);
			system->status&=~SSTATUS_PAGESHOW;
		}
		else
		{
			obj->status|=OSTATUS_REDRAW;
			system->status|=SSTATUS_REDRAW;
		}
		break;
	}
	return PFPrintEscape(gstate,FALSE);
}

Defines defines[]={
{ "INK-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &ink,           &Kink },
{ "PAPER-COLOUR",   PW_DEFINE_COLOUR,   NULL,   &paper,         &Kpaper },
{ "FONT",           PW_DEFINE_FONT,     NULL,   loose_font,     &recalculate },
{ "FONTSIZE",       PW_DEFINE_PT,       NULL,   &loose_fontsize,&recalculate },
{ "MAXLENGTH",      PW_DEFINE_SHORT,    NULL,   &bufflen,       NULL },
{ "ITEMWIDTH",      PW_DEFINE_PT,       NULL,   &xsiz,          NULL },
{ 0 }
};

Error Define(char *name, char *value)
{
	return PWHandleDefine("EDLINE-",defines,name,value);
}
	
Error DetermineSize(PWObject object)
{
	PWRegion *region=object->event.region;
	Object *obj=(Object *)object;
	Gstate gstate=obj->event.region->system->gstate;

	if (!gstate) return ERR_OK;

	obj->offset=0;					/* make as much visible as possible */

	if (recalculate) Calculate(gstate,(PWObject)obj);
	region->tot_ysiz=short2pt(PF2ypix(loose_height));
	region->tot_xsiz=short2pt(obj->xsiz);

	region->hit_off_lft=region->hit_off_rgt=
	region->hit_off_top=region->hit_off_bot=0;

	return ERR_OK;
}


PWTypeDef init={
	PW_TYPEDEF_FLAG, TYPE, PW_TYPE_EVENT_REGION,
	&Create, &Change, NULL, &Remove, NULL, NULL, &Handle, &Define,
	&Draw, &DetermineSize, NULL
};
