/*
	ProWesS, PROGS Windowing System

	edit line of text type for ProWesS, "EDLI"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 16 July 1995
	26.10.1995 fixed problem with not restoring PFPrintEscape in time
	18.12.1995 now also use Draw routine from PW_TYPE_iTEM
	23.03.1996 adjusted for internal use of pixel values
	11.04.1996 use PWHandleDefine && use ProWesS defaults
	07.05.1996 keypress now cleared when done
	11.05.1996 PW_EDLINE_ACTIVATE added
	07.06.1996 PW_EDLINE_ACTION_TEST added, including three default actions
	02.07.1996 PW_EDLINE_ACTION_TEST routine now also called before first key
*/

#include "mem_h"
#include "str_h"
#include "ctype_h"
#include "io_h"

#include "PWHandler_h"

#define TYPE PW_TYPE_EDLINE

/* global statusses */

#define p100 short2pt(100)
ColourRGB ink	= { 0,    0,    0    }; char Kink  =FALSE;
ColourRGB paper = { p100, p100, p100 }; char Kpaper=FALSE;
char loose_font[PF_MAXFONTNAME]=""; /* name of PROforma font to use */
pt loose_fontsize=0;				/* fontsize to use */
pt loose_baseline=0;				/* font baseline position (at fontsize) */
pt loose_height=0;					/* height of font (at fontsize) */
short bufflen=256;					/* default maximum size of strings */
pt xsiz=short2pt(120);				/* default minimum size of edline object */
char recalculate=TRUE;				/* set when baseline and height should be determined */

PWTypeDef *border;

/* loose item object definition */

typedef struct _Object {
	PWBaseObject;
	int bufflen;			/* length of buffer */
	pix xsiz;				/* minimum width of item */
	PWObject keypress;		/* keypress attached to this object */
	Error (*DoAfter)(PWObject object);	/* call after pressing <enter> */
	Error (*DoAction)(PWObject object); /* action for when DO on edline */
	Error (*ActionTest)(PWObject object, EdlineInfo *info);
	struct _Object *next, *prev, *up, *down, *after;
	char string[2]; 		/* string to edit - length==bufflen */
	} Object;


/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Error err;
	short buffer=bufflen;

	LinkPROforma();

	err=PWGetType(PW_TYPE_iTEM, &border);
	if (err) return err;

	/* see if there is a PW_EDLINE_MAXLENGTH tag */
	while (*taglist)
	{
		if (*taglist==PW_EDLINE_MAXLENGTH)
		{
			buffer=(*(taglist+1));
			*taglist&=PW_TAGMASK_SKIP;
		}
		taglist=PWSkipTag(taglist);
	}

	/* allocate specific object - which should be cleared (automatic) */
	err=PWAllocate(event.region->system,sizeof(Object)+buffer,ret);
	if (err) return err;

	/* make sure the item is available */
	(*(Object **)ret)->bufflen=buffer;
	(*(Object **)ret)->xsiz=PF2xpix(xsiz);
	event.region->wait_event=PW_EVENT_HIT|PW_EVENT_DO;
	return err;
}

Error Remove(PWObject object)
{
	Object *obj=(Object *)object;
	if (obj->keypress) PWRemove(obj->keypress);
	return PWRelease((PWObject)obj,(PWObject)obj);
}

Error Draw(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *reg=obj->event.region;
	Gstate gstate=reg->system->gstate;
	pix hit_xsiz=reg->hit_xend-reg->hit_xorg;
	pix hit_ysiz=reg->hit_yend-reg->hit_yorg;

	if (!gstate) return ERR_OK;

	PFPWindowSubSet(gstate,hit_xsiz,hit_ysiz,reg->hit_xorg,reg->hit_yorg);
	PFPaperColourRGB(gstate,&paper);
	PFColourRGB(gstate,&ink);
	PFPageClear(gstate);
	PFFontLoad(gstate,loose_font);
	PFFontScale(gstate,loose_fontsize);
	PFMoveTo(gstate,0,(ypix2PF(hit_ysiz)-loose_height)/2+loose_baseline);
	PFStringShow(gstate,obj->string);
	PFWindowSubRestore(gstate);

	return border->PWDraw((PWObject)obj);
}

Error Entry(PWObject object)
{
	return border->PWRegionEntry(object);
}

Error Exit(PWObject object)
{
	return border->PWRegionExit(object);
}

Error EditLine(Object *obj)
{
	PWSystem *system=obj->event.region->system;
	Gstate gstate=system->gstate;
	Object *nextobj=NULL;

	if (!gstate) return ERR_OK;

	PFPaperColourRGB(gstate,&paper);
	PFColourRGB(gstate,&ink);
	PFFontLoad(gstate,loose_font);
	PFFontScale(gstate,loose_fontsize);
	PFLineWidth(gstate,0);
	PFPathStroked(gstate);
	PFPrintEscape(gstate,TRUE);

	while (obj)
	{
		pt curpos=0;					/* cursor position, in window */
		pt offset=0;					/* x pageorigin where string was displayed */
		PWRegion *reg=obj->event.region;
		pix hit_xsiz=reg->hit_xend-reg->hit_xorg;
		pix hit_ysiz=reg->hit_yend-reg->hit_yorg;
		pt min_cur, max_cur, dist;
		pt ypos=(ypix2PF(hit_ysiz)-loose_height)/2;
		char w[2]={0,0}; pt cw;
		EdlineInfo info;
		Error err=ERR_OK;
		char starting=TRUE;

		info.maxlength=obj->bufflen;
		info.string=obj->string;
		info.length=strlen(obj->string);
		info.cursor=info.length;
		info.key=0;

		PFPWindowSubSet(gstate,hit_xsiz,hit_ysiz,reg->hit_xorg,reg->hit_yorg);
		
		/* next two lines are ridiculous, but putting it in one line seems to */
		/* make c68 (or 68030) choke. The code looks ok, but doesn't work. So ... */
/*		  max_cur=-min(loose_fontsize,hit_xsiz/8); */
/*		  max_cur+=hit_xsiz; */
		min_cur=min(loose_fontsize,xpix2PF(hit_xsiz)/8);
		max_cur=xpix2PF(hit_xsiz)-min_cur;

		PFStringWidth(gstate,obj->string,&curpos);

		while(info.key!='\n' && !nextobj)
		{
			/* make sure cursor is visible in window */
			if ( (dist=curpos-max_cur-offset)>0 ||
				 (dist=curpos-min_cur-offset)<0 )
			{
				PFPageOriginSet(gstate,dist,0);
				offset+=dist;
				if (offset<0)
				{
					PFPageOriginSet(gstate,-offset,0);
					offset=0;
				}
			}
			PFPageClear(gstate);
			PFMoveTo(gstate,0,ypos+loose_baseline);
			PFStringShow(gstate,obj->string);
			PFMoveTo(gstate,curpos,ypos);
			PFLineR(gstate,0,loose_height);
			PFPathDraw(gstate);
			PFPageShow(gstate);

			/* get next character */
			if (!starting) IOGet(system->window,&info.key);
			starting=FALSE;

			/* handle the keypress, allow callback mechanism to check */
			/* whether the keypress is valid */
			/* The callback routine can have three possible results */
			/* ACCEPT keypress is valid, use normal processing */
			/* NO ACCEPT do not accept */
			/* CHANGE do not use normal processing, the string and/or */
			/*		cursor position are changed by the callback routine */
			if (obj->ActionTest)
				err=obj->ActionTest((PWObject)obj,&info);

			if (err)
			{
				if (err==PW_EDLINE_TEST_CHANGE)
				{
					/* string and/or cursor position have changed */
					/* adjust internal structures */
					/* determine new value of currpos */
					int i;
					curpos=0;
					for (i=0 ; i<info.cursor ; i++)
					{
						w[0]=obj->string[i];
						PFStringWidth(gstate,w,&cw);
						curpos+=cw;
					}
				}
				continue;
			}

			/* and put it in the string */
			if (isprint(info.key) && info.length<obj->bufflen-1)
			{
				/* put it in the string */
				char *pos=&obj->string[info.cursor];
				char *end=&obj->string[++info.length];
				while (end>=pos) { *(end+1)=*end; end--; }
				*pos=info.key;
				info.cursor++;
				w[0]=info.key;
				PFStringWidth(gstate,w,&cw);
				curpos+=cw;
			}
			else switch(info.key)
			{
			case '\n':
				nextobj=obj->after;
				break;
			case '\t':
				nextobj=obj->next;
				break;
			case (char)0xfd:
				nextobj=obj->prev;
				break;
			case (char)0xd0:
				nextobj=obj->up;
				break;
			case (char)0xd8:
				nextobj=obj->down;
				break;
			case (char)0xc0 : /* left */
				if (info.cursor)
				{
					info.cursor--;
					w[0]=obj->string[info.cursor];
					PFStringWidth(gstate,w,&cw);
					curpos-=cw;
				}
				break;
			case (char)0xc1 : /* alt left - cursor to start of line */
				info.cursor=0;
				curpos=0;
				break;
			case (char)0xc2 : /* ctrl left - backspace */
				if (info.cursor)
				{
					char *pos=&obj->string[info.cursor];
					w[0]=*(pos-1);
					info.cursor--;
					PFStringWidth(gstate,w,&cw);
					curpos-=cw;
					while (*pos) { *(pos-1)=*pos; pos++; }
					*(pos-1)='\0';
					info.length--;
				}
				break;
			case (char)0xc3 : /* ctrl alt left - delete start of line */
				{
					char *pos=obj->string;
					int i=info.length-info.cursor;
					for(;i;i--) *pos++=*(pos+info.cursor);
					*pos=0;
					info.length-=info.cursor;
					info.cursor=0;
					curpos=0;
				}
				break;
			case (char)0xc4 : /* shift left */
				{
					int gotspace=obj->string[info.cursor-1]==' ';
					while ( info.cursor &&
							( (obj->string[info.cursor-1]==' ' &&  gotspace) ||
							  (obj->string[info.cursor-1]!=' ' && !gotspace) ) )
					{
						info.cursor--;
						w[0]=obj->string[info.cursor];
						PFStringWidth(gstate,w,&cw);
						curpos-=cw;
						if (gotspace) gotspace=obj->string[info.cursor-1]==' ';
					}
				}
				break;
			case (char)0xc6 : /* shift ctrl left */
				{
					int oldcursor=info.cursor;
					int gotspace=obj->string[info.cursor-1]==' ';
					while ( info.cursor &&
							( (obj->string[info.cursor-1]==' ' &&  gotspace) ||
							  (obj->string[info.cursor-1]!=' ' && !gotspace) ) )
					{
						info.cursor--;
						w[0]=obj->string[info.cursor];
						PFStringWidth(gstate,w,&cw);
						curpos-=cw;
						if (gotspace) gotspace=obj->string[info.cursor-1]==' ';
					}
					if (info.cursor!=oldcursor)
					{
						char *pos=&obj->string[info.cursor];
						oldcursor-=info.cursor;
						while (*pos++=*(pos+oldcursor));
						info.length-=oldcursor;
					}
				}
				break;
			case (char)0xc8 : /* right */
				if (info.cursor<info.length)
				{
					w[0]=obj->string[info.cursor];
					PFStringWidth(gstate,w,&cw);
					info.cursor++;
					curpos+=cw;
				}
				break;
			case (char)0xc9 : /* alt right */
				info.cursor=info.length;
				PFStringWidth(gstate,obj->string,&curpos);
				break;
			case (char)0xca : /* ctrl right */
				{
					char *pos=&obj->string[info.cursor];
					while (*pos++=*(pos+1));
					info.length--;
				}
				break;
			case (char)0xcb : /* ctrl alt right */
				obj->string[info.cursor]=0;
				info.length=strlen(obj->string);
				break;
			case (char)0xcc : /* shift right */
				{
					int gotspace=obj->string[info.cursor]==' ';
					while ( info.cursor<info.length &&
							( (obj->string[info.cursor]==' ' &&  gotspace) ||
							  (obj->string[info.cursor]!=' ' && !gotspace) ) )
					{
						w[0]=obj->string[info.cursor];
						PFStringWidth(gstate,w,&cw);
						info.cursor++;
						curpos+=cw;
						if (!gotspace) gotspace=obj->string[info.cursor]==' ';
					}
				}
				break;
			case (char)0xce : /* shift ctrl right */
				{
					int gotspace=obj->string[info.cursor]==' ';
					int endpos=info.cursor;
					while ( endpos<info.length &&
							( (obj->string[endpos]==' ' &&	gotspace) ||
							  (obj->string[endpos]!=' ' && !gotspace) ) )
					{
						endpos++;
						if (!gotspace) gotspace=obj->string[endpos]==' ';
					}
					if (info.cursor!=endpos)
					{
						char *pos=&obj->string[info.cursor];
						endpos-=info.cursor;
						while (*pos++=*(pos+endpos));
						info.length-=endpos;
					}
				}
				break;
			case (char)0xda : /* ctrl down */
				obj->string[0]='\0';
				curpos=info.length=info.cursor=0;
				break;
			}
		}

		PFWindowSubRestore(gstate);
		PFPrintEscape(gstate,FALSE);

		/* call post-processing routine (if any) */
		if (obj->DoAfter) obj->DoAfter((PWObject)obj);

		Draw((PWObject)obj);
		PFPageShow(gstate);
		system->status&=~SSTATUS_PAGESHOW;

		obj=nextobj;
		nextobj=NULL;
	}
	/* clear keypress !! */
	{
		PointerInfo pr;
		short xp, yp;
		Error err;

		err=WINPointerRead(system->window,1,EVENT_KEYUP,&xp,&yp,&pr);
		if (err=ERR_NC && pr.event&(EVENT_KEYDOWN|EVENT_KEYPRESS))
			WINPointerRead(system->window,FOREVER,EVENT_KEYUP,&xp,&yp,&pr);
	}
	return ERR_OK;
}

Error Handle(PWObject object, PWWait event)
{
	Object *obj=(Object *)object;

	if (event==PW_EVENT_DO && obj->DoAction)
		return obj->DoAction((PWObject)obj);

	if (event==PW_EVENT_HIT || event==PW_EVENT_DO)
		return EditLine(obj);
	else
		return ERR_OK;
}

static Error natural(PWObject object, EdlineInfo *info)
{
	if (isprint(info->key) && !isdigit(info->key))
		return PW_EDLINE_TEST_NOACCEPT;
	else
		return PW_EDLINE_TEST_ACCEPT;
}

static Error integer(PWObject object, EdlineInfo *info)
{
	if (info->key=='+' || info->key=='-')
	{
		if (info->string[0]=='+' || info->string[0]=='-')
		{
			info->string[0]=info->key;
			return PW_EDLINE_TEST_CHANGE;
		}
		if (info->length+1<info->maxlength)
		{
			MEMCopy(info->maxlength-1,info->string,info->string+1);
			info->string[0]=info->key;
			info->cursor++;
			info->length++;
			return PW_EDLINE_TEST_CHANGE;
		}
		return PW_EDLINE_TEST_NOACCEPT;
	}
	if (isprint(info->key) && !isdigit(info->key))
		return PW_EDLINE_TEST_NOACCEPT;
	else
		return PW_EDLINE_TEST_ACCEPT;
}

static Error number(PWObject object, EdlineInfo *info)
{
	if (info->key=='+' || info->key=='-')
	{
		if (info->string[0]=='+' || info->string[0]=='-')
		{
			info->string[0]=info->key;
			return PW_EDLINE_TEST_CHANGE;
		}
		if (info->length+1<info->maxlength)
		{
			MEMCopy(info->maxlength-1,info->string,info->string+1);
			info->string[0]=info->key;
			info->cursor++;
			info->length++;
			return PW_EDLINE_TEST_CHANGE;
		}
		return PW_EDLINE_TEST_NOACCEPT;
	}
	if (info->key=='.')
	{
		/* only accept if no decimal point yet */
		char *str;
		for (str=info->string ; *str ; str++)
			if (*str=='.') return PW_EDLINE_TEST_NOACCEPT;
		return PW_EDLINE_TEST_ACCEPT;
	}
	if (isprint(info->key) && !isdigit(info->key))
		return PW_EDLINE_TEST_NOACCEPT;
	else
		return PW_EDLINE_TEST_ACCEPT;
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	Gstate gstate=system->gstate;
	char unrecognised_tag=FALSE;

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_EDLINE_SET:
			STRCopySafe((char *)(*(taglist+1)),obj->string,obj->bufflen);
			obj->status|=OSTATUS_REDRAW;
			system->status|=SSTATUS_REDRAW;
			break;
		case PW_EDLINE_GET:
			/* two parameters : buffer length and buffer base */
			STRCopySafe(obj->string,(char *)(*(taglist+2)),(*(taglist+1)));
			break;
		case PW_EDLINE_WIDTH:
			obj->xsiz=PF2xpix(*(taglist+1));
			system->status|=SSTATUS_WINCHANGE;
			break;
		case PW_EDLINE_WIDTH_PIX:
			obj->xsiz=(*(taglist+1));
			system->status|=SSTATUS_WINCHANGE;
			break;
		case PW_EDLINE_WIDTH_FS:
			obj->xsiz=PF2xpix(fixmul((pt)(*(taglist+1)),loose_fontsize));
			system->status|=SSTATUS_WINCHANGE;
			break;
		case PW_EDLINE_KEYPRESS:
			/* attach keypress */
			if (!obj->keypress)
			{
				err=PWCreate(object,&obj->keypress,PW_TYPE_KEYPRESS,NULL);
				if (err) return err;
			}
			err=PWChange(obj->keypress,PW_KEYPRESS,(*(taglist+1)),NULL);
			if (err) return err;
			break;
		case PW_EDLINE_ACTION_AFTER:
			obj->DoAfter=(Error (*)(PWObject))(*(taglist+1));
			break;
		case PW_EDLINE_ACTION_DO:
			obj->DoAction=(Error (*)(PWObject))(*(taglist+1));
			break;
		case PW_EDLINE_ACTION_TEST:
			if ((*(taglist+1))<0)
			{
				switch (*(taglist+1))
				{
				case PW_EDLINE_ACTION_TEST_NATURAL:
					obj->ActionTest=natural;
					break;
				case PW_EDLINE_ACTION_TEST_INTEGER:
					obj->ActionTest=integer;
					break;
				case PW_EDLINE_ACTION_TEST_FLOAT:
					obj->ActionTest=number;
					break;
				}
			}
			else
				obj->ActionTest=(Error (*)(PWObject, EdlineInfo *))(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_AFTER:
			obj->after=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_NEXT:
			obj->next=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_PREV:
			obj->prev=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_UP:
			obj->up=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_DOWN:
			obj->down=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_ACTIVATE:
			if (system->activated) EditLine(obj);
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Defines defines[]={
{ "INK-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &ink,           &Kink },
{ "PAPER-COLOUR",   PW_DEFINE_COLOUR,   NULL,   &paper,         &Kpaper },
{ "FONT",           PW_DEFINE_FONT,     NULL,   loose_font,     &recalculate },
{ "FONTSIZE",       PW_DEFINE_PT,       NULL,   &loose_fontsize,&recalculate },
{ "MAXLENGTH",      PW_DEFINE_SHORT,    NULL,   &bufflen,       NULL },
{ "ITEMWIDTH",      PW_DEFINE_PT,       NULL,   &xsiz,          NULL },
{ 0 }
};

Error Define(char *name, char *value)
{
	return PWHandleDefine("EDLINE-",defines,name,value);
}

Error DetermineSize(PWObject object)
{
	PWRegion *region=object->event.region;
	Object *obj=(Object *)object;
	Gstate gstate=obj->event.region->system->gstate;

	if (!gstate) return ERR_OK;

	if (!Kink) PWQuery(object,PW_DEFAULT_FOREGROUND,&ink);
	if (!Kpaper) PWQuery(object,PW_DEFAULT_BACKGROUND,&paper);
	if (recalculate)
	{
		Bbox bbox;
		recalculate=FALSE;
		if (!loose_fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&loose_fontsize);
		if (!*loose_font) PWQuery(object,PW_DEFAULT_FONT,loose_font);
		PFFontLoad(gstate,loose_font);
		PFFontScale(gstate,loose_fontsize);
		PFFontBbox(gstate,&bbox);
		loose_baseline=bbox.yorg;
		loose_height=bbox.ysiz;
	}
	region->tot_ysiz=short2pt(PF2ypix(loose_height));
	region->tot_xsiz=short2pt(obj->xsiz);

	return border->PWDetermineSize((PWObject)obj);
}


PWTypeDef init={
	PW_TYPEDEF_FLAG, TYPE, PW_TYPE_EVENT_REGION,
	&Create, &Change, NULL, &Remove, &Entry, &Exit, &Handle, &Define,
	&Draw, &DetermineSize, NULL
};
