/*
	ProWesS, PROGS Windowing System

	Info string type "INFS", the info string is always completely visible

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 23 August 1995
	14.10.1995 width expanded by 2 pixels and show at 1 pixel from side
	22.03.1996 adjusted for internal use of pixel coordinates
	11.04.1996 use PWHandleDefine && use ProWesS defaults
*/

#include "mem_h"

#include "PWHandler_h"

#define TYPE PW_TYPE_INFOSTRING

/* global statusses */

#define p100 short2pt(100)
ColourRGB ink	= { 0,    0,    0    }; char Kink  =FALSE;
ColourRGB paper = { p100, p100, p100 }; char Kpaper=FALSE;

char font[PF_MAXFONTNAME]="";		/* name of PROforma font to use */
pt fontsize=0;						/* fontsize to use */
pt baseline=0;						/* font baseline position (at fontsize) */
pt height=0;						/* height of font (at fontsize) */
char recalculate=TRUE;

/* loose item object definition */

typedef struct {
	PWBaseObject;
	char *item; 					/* item identifier */
	char fixedsize; 				/* inverse of autosize status */
	} Object;

/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	LinkPROforma();
	/* allocate specific object - which should be cleared (automatic) */
	return PWAllocate(event.region->system,sizeof(Object),ret);
}

Error Remove(PWObject object)
{
	return PWRelease(object,object);
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_INFOSTRING_TEXT :
			obj->item=(char *)(*(taglist+1));
			if (!obj->fixedsize)
				system->status|=SSTATUS_WINCHANGE;
			else
			{
				obj->status|=OSTATUS_REDRAW;
				system->status|=SSTATUS_REDRAW;
			}
			break;
		case PW_INFOSTRING_AUTOSIZE :
			if (*(taglist+1))
				obj->fixedsize=FALSE;
			else
				obj->fixedsize=TRUE;
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Error Draw(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *reg=obj->event.region;
	Gstate gstate=reg->system->gstate;
	pix hit_xsiz=reg->hit_xend-reg->hit_xorg;
	pix hit_ysiz=reg->hit_yend-reg->hit_yorg;
	int i;
	char *item=obj->item;

	if (!gstate && !item) return ERR_OK;

	PFPWindowSubSet(gstate,hit_xsiz,hit_ysiz,reg->hit_xorg,reg->hit_yorg);
	PFPaperColourRGB(gstate,&paper);
	PFPageClear(gstate);
	PFColourRGB(gstate,&ink);
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	PFExtraEOS(gstate,'\n');
	for(i=0 ; *item ; i++)
	{
		PFMoveTo(gstate,xpix2PF(1),i*height+baseline);
		PFStringShow(gstate,item);
		while (*item && *item++!='\n');
	}

	PFWindowSubRestore(gstate);
	PFExtraEOS(gstate,'\0');

	return ERR_OK;
}

Defines defines[]={
{ "INK-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &ink,       &Kink },
{ "PAPER-COLOUR",   PW_DEFINE_COLOUR,   NULL,   &paper,     &Kpaper },
{ "FONT",           PW_DEFINE_FONT,     NULL,   font,       &recalculate },
{ "FONTSIZE",       PW_DEFINE_PT,       NULL,   &fontsize,  &recalculate },
{ 0 }
};

Error Define(char *name, char *value)
{
	return PWHandleDefine("INFOSTRING-",defines,name,value);
}

Error DetermineSize(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *region=obj->event.region;
	Gstate gstate=region->system->gstate;
	char *item=obj->item;
	pt itemwidth, itemheight;
	pt width;

	region->hit_off_lft=region->hit_off_rgt=
	region->hit_off_top=region->hit_off_bot=0;

	if (!gstate) return ERR_OK;

	if (!Kink) PWQuery(object,PW_DEFAULT_FOREGROUND,&ink);
	if (!Kpaper) PWQuery(object,PW_DEFAULT_BACKGROUND,&paper);
	if (recalculate)
	{
		Bbox bbox;
		recalculate=FALSE;
		if (!fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&fontsize);
		if (!*font) PWQuery(object,PW_DEFAULT_FONT,font);
		PFFontLoad(gstate,font);
		PFFontScale(gstate,fontsize);
		PFFontBbox(gstate,&bbox);
		baseline=bbox.yorg;
		height=bbox.ysiz;
	}

	/* set itemwidth */
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	PFExtraEOS(gstate,'\n');
	itemheight=0;
	itemwidth=0;
	if (item) while(*item)
	{
		itemheight+=height;
		PFStringWidth(gstate,item,&width);
		if (width>itemwidth) itemwidth=width;
		while (*item && *item++!='\n');
	}
	PFExtraEOS(gstate,'\0');

	region->tot_ysiz=short2pt(PF2ypix(itemheight));
	region->tot_xsiz=short2pt(PF2xpix(itemwidth)+2);

	return ERR_OK;
}


PWTypeDef init={
	PW_TYPEDEF_FLAG, TYPE, PW_TYPE_EVENT_REGION,
	&Create, &Change, NULL, &Remove, NULL, NULL, NULL, &Define,
	&Draw, &DetermineSize, NULL
};
