/*
	ProWesS, PROGS Windowing System

	Info text type "INFT", the info text is contained in an applic window

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 24 August 1995
	14.10.1995 made sure width always expanded by 2 pixels (and show at x1pix)
	11.04.1996 use PWHandleDefine && use ProWesS defaults
*/

#include "mem_h"

#include "PWHandler_h"

#define TYPE PW_TYPE_INFOTEXT

/* global statusses */

#define p100 short2pt(100)
ColourRGB ink	= { 0,    0,    0    }; char Kink  =FALSE;
ColourRGB paper = { p100, p100, p100 }; char Kpaper=FALSE;

char font[PF_MAXFONTNAME]="";		/* name of PROforma font to use */
pt fontsize=0;						/* fontsize to use */
pt baseline=0;						/* font baseline position (at fontsize) */
pt height=0;						/* height of font (at fontsize) */
char recalculate=TRUE;

/* loose item object definition */

typedef struct {
	PWBaseObject;
	PWObject applic;				/* applic object connected with infotext */
	PWTypeDef *atype;				/* type definition for applic */
	char *item; 					/* item identifier */
	} Object;

Error Remove(PWObject object);		/* forward declaration */

static void Calculate(Gstate gstate, PWObject object)
{
	Bbox bbox;

	if (!Kink) PWQuery(object,PW_DEFAULT_FOREGROUND,&ink);
	if (!Kpaper) PWQuery(object,PW_DEFAULT_BACKGROUND,&paper);
	if (!fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&fontsize);
	if (!*font) PWQuery(object,PW_DEFAULT_FONT,font);

	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	PFFontBbox(gstate,&bbox);
	baseline=bbox.yorg;
	height=bbox.ysiz;
	recalculate=FALSE;
}

/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Object *obj;
	Error err;
	int null=0;

	LinkPROforma();

	/* allocate specific object - which should be cleared (automatic) */
	err=PWAllocate(event.region->system,sizeof(Object),ret);
	if (err) return err;

	obj=(Object *)*ret;
	if (recalculate) Calculate(event.region->system->gstate,(PWObject)obj);

	/* initialise object */
	err=PWGetType(PW_TYPE_APPLIC, &obj->atype);
	if (err) { Remove((PWObject)obj); return err; }

	err=obj->atype->PWCreate((PWObject)obj,&obj->applic,event,&null);
	if (err) { Remove((PWObject)obj); return err; }
	/* fill in the base object (normally done by system, which is bypassed) */
	obj->applic->type=obj->atype;
	obj->applic->event=event;

	return err;
}

Error Remove(PWObject object)
{
	Object *obj=(Object *)object;

	/* remove applic */
	if (obj->applic)
		obj->atype->PWRemove(obj->applic);

	return PWRelease((PWObject)obj,(PWObject)obj);
}

Error CanvasDraw(PWObject object, CanvasInfo *rinfo)
{
	Object *obj=(Object *)object;
	Gstate gstate=rinfo->gstate;
	int i;
	char *item;
	int first,last;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);

	if (obj->item==NULL) return ERR_OK;

	if (recalculate) Calculate(gstate,(PWObject)obj);

	PFPaperColourRGB(gstate,&paper);
	PFPageClear(gstate);
	PFColourRGB(gstate,&ink);
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	PFExtraEOS(gstate,'\n');
	PFPageOriginSet(gstate,rinfo->xorg,rinfo->yorg*height);

	item=obj->item;
	first=rinfo->yorg;
	last =rinfo->ysiz/height+first+1;
	for(i=0 ; *item ; i++)
	{
		if (i>=first)
		{
			PFMoveTo(gstate,xpix2PF(1),i*height+baseline);
			PFStringShow(gstate,item);
		}
		if (i==last) break;
		while (*item && *item++!='\n');
	}

	PFExtraEOS(gstate,'\0');

	return ERR_OK;
}

Error calcsize(PWObject object, pt *size)
{
	if (recalculate) Calculate(object->event.region->system->gstate,object);
	*size=*size/height;
	return ERR_OK;
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;
	Gstate gstate=system->gstate;

	if (recalculate) Calculate(gstate,(PWObject)obj);

	if (!(obj->status&OSTATUS_INITIALISED))
	{
		obj->status|=OSTATUS_INITIALISED;
		PWChange(obj->applic,
					PW_APPLIC_CANVAS_LIST,
						PW_OBJECT_AUXILIARY, obj,
						PW_CANVAS_YSIZE, height,
						PW_CANVAS_ACTION_REDRAW, &CanvasDraw,
						NULL,
					PW_APPLIC_SCROLLBAR_Y,
					PW_APPLIC_YSCROLL_LIST,
						PW_OBJECT_AUXILIARY, obj,
						PW_SCROLL_MINDIST, 1,
						PW_SCROLL_MAXDIST, -1,
						PW_SCROLL_CALCSIZE, &calcsize,
						NULL,
					PW_APPLIC_SCROLLBAR_X,
					PW_APPLIC_XSCROLL_LIST,
						PW_OBJECT_AUXILIARY, obj,
						PW_SCROLL_MINDIST, fontsize,
						PW_SCROLL_MAXDIST, -fontsize,
						NULL,
					NULL);
	}

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_INFOTEXT_TEXT :
			{
				char *item=(char *)(*(taglist+1));
				pt itemwidth, width;
				int itemheight;

				obj->item=item;

				/* set itemwidth */
				PFFontLoad(gstate,font);
				PFFontScale(gstate,fontsize);
				PFExtraEOS(gstate,'\n');
				itemheight=0;
				itemwidth=0;
				if (item) while(*item)
				{
					itemheight++;
					PFStringWidth(gstate,item,&width);
					width+=xpix2PF(2);
					if (width>itemwidth) itemwidth=width;
					while (*item && *item++!='\n');
				}
				PFExtraEOS(gstate,'\0');

				PWChange(obj->applic,
							PW_APPLIC_YSCROLL_LIST,
								PW_SCROLL_MAXIMUM, itemheight,
								NULL,
							PW_APPLIC_XSCROLL_LIST,
								PW_SCROLL_MAXIMUM, itemwidth,
								NULL,
							PW_APPLIC_CANVAS_LIST,
								PW_CANVAS_REDRAW,
								NULL,
							NULL);
			}
			break;
		case PW_INFOTEXT_WIDTH:
			PWChange(obj->applic,
					PW_APPLIC_CANVAS_LIST,
						PW_CANVAS_XSIZE, (*(taglist+1)),
						NULL,
					NULL);
			break;
		case PW_INFOTEXT_WIDTH_PIX:
			PWChange(obj->applic,
					PW_APPLIC_CANVAS_LIST,
						PW_CANVAS_XSIZE, xpix2PF(*(taglist+1)),
						NULL,
					NULL);
			break;
		case PW_INFOTEXT_WIDTH_FS:
			PWChange(obj->applic,
					PW_APPLIC_CANVAS_LIST,
						PW_CANVAS_XSIZE, fixmul((*(taglist+1)),fontsize),
						NULL,
					NULL);
			break;
		case PW_INFOTEXT_LINES:
			PWChange(obj->applic,
					PW_APPLIC_CANVAS_LIST,
						PW_CANVAS_YSIZE, (*(taglist+1))*height,
						NULL,
					NULL);
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Defines defines[]={
{ "INK-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &ink,       &Kink },
{ "PAPER-COLOUR",   PW_DEFINE_COLOUR,   NULL,   &paper,     &Kpaper },
{ "FONT",           PW_DEFINE_FONT,     NULL,   font,       &recalculate },
{ "FONTSIZE",       PW_DEFINE_PT,       NULL,   &fontsize,  &recalculate },
{ 0 }
};

Error Define(char *name, char *value)
{
	return PWHandleDefine("INFOTEXT-",defines,name,value);
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, TYPE, PW_TYPE_EVENT_REGION,
	&Create, &Change, NULL, &Remove, NULL, NULL, NULL, &Define,
	NULL, NULL, NULL
};
