/*
	ProWesS, PROGS Windowing System

	item support type "iTEM"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 January 1995
	18.12.1995 added Draw routine to have equivalent type without active border
	21.03.1996 changed internal data to more pixels
	06.06.1996 use system default ink when none defined
	28.06.1996 faster (and shorter) border drawing thanks to PFP and better
			   handling of linewidth in PROforma
*/

#include "mem_h"

#include "PWHandler_h"

/* global statusses */

#define p100 short2pt(100)
pt Xpt2pix, Ypt2pix;
pt width, xwidth=pt_one,ywidth=pt_one;	/* width nominal and each direction */
ColourRGB ink	= { p100, 0, 0 }; char Kink  =FALSE;
ColourRGB paper = { 0,    0, 0 }; char Kpaper=FALSE;

static Error DrawBorder(PWObject obj, ColourRGB *colour)
{
	PWRegion *reg=obj->event.region;
	Gstate gstate=reg->system->gstate;
	pt xorg=short2pt(reg->hit_xorg);
	pt yorg=short2pt(reg->hit_yorg);
	pt xsiz=short2pt(reg->hit_xend)-xorg+xwidth;
	pt ysiz=short2pt(reg->hit_yend)-yorg+ywidth;

	PFPathStroked(gstate);
	PFColourRGB(gstate, colour);
	PFLineWidth(gstate, width);

	PFPMoveTo(gstate, xorg-xwidth/2, yorg-ywidth/2);
	PFPLineR(gstate, 0, ysiz);
	PFPLineR(gstate, xsiz, 0);
	PFPLineR(gstate, 0, -ysiz);
	PFPathClose(gstate);
	PFPathDraw(gstate);

	reg->system->status|=SSTATUS_PAGESHOW;

	return ERR_OK;
}

Error Entry(PWObject object)
{
	return DrawBorder(object,&ink);
}

Error Exit(PWObject object)
{
	return DrawBorder(object,&paper);
}

Error Draw(PWObject objetc)
{
	return ERR_OK;
}

Error Define(char *name, char *value)
{
	ColourRGB *colour;
	double red, green, blue, db;

	LinkPROforma();
	PWpoint2pixel(&Xpt2pix,&Ypt2pix);

	if (STRSameCD(name,"ITEM-BORDER-COLOUR"))
	{
		Kink=TRUE;
		colour=&ink;
readcolour:
		STRFormatGet(value,NULL,"%f %f %f",&red,&green,&blue);
		colour->red  =double2pt(red);
		colour->green=double2pt(green);
		colour->blue =double2pt(blue);
		return ERR_OK;
	}
	if (STRSameCD(name,"SYSTEM-BACKGROUND-COLOUR"))
	{
		Kpaper=TRUE;
		colour=&paper;
		goto readcolour;
	}
	if (STRSameCD(name,"ITEM-BORDER-WIDTH"))
	{
		STRFormatGet(value,NULL,"%f",&db);
		width=double2pt(db);

		xwidth=pt_round(fixmul(width,Xpt2pix)); if (xwidth<pt_one) xwidth=pt_one;
		ywidth=pt_round(fixmul(width,Ypt2pix)); if (ywidth<pt_one) ywidth=pt_one;
		return ERR_OK;
	}
	return ERR_IPAR;
}

Error DetermineSize(PWObject obj)
{
	PWRegion *region=obj->event.region;

	if (!Kink)	 PWQuery(obj,PW_DEFAULT_MIDDLEGROUND,&ink);
	if (!Kpaper) PWQuery(obj,PW_DEFAULT_BACKGROUND,&paper);

	region->hit_off_lft=region->hit_off_rgt=pt2short(xwidth);
	region->hit_off_top=region->hit_off_bot=pt2short(ywidth);

	region->tot_xsiz+=2*xwidth;
	region->tot_ysiz+=2*ywidth;
	return ERR_OK;
}


PWTypeDef init={
	PW_TYPEDEF_FLAG, PW_TYPE_iTEM, PW_TYPE_EVENT_REGION,
	NULL, NULL, NULL, NULL, Entry, Exit, NULL, Define,
	Draw, DetermineSize, NULL
};
