/*
	ProWesS, PROGS Windowing System

	Keypress object "KEY "

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 19 February 1995
*/

#include "mem_h"

#include "ProWesS_h"
#include "PWTypeDef_h"
#include "PWHandler_h"

/* global statusses */

/* object definition */
typedef struct {
	PWBaseObject;					/* common start for all objects */
	PWObject parent;				/* parent for keypress */
	Error (*Action)(PWObject object);	/* action for keypress */
	} Object;


/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Error err;

	/* allocate specific object - which should be cleared (automatic) */
	err=PWAllocate(event.region->system,sizeof(Object),ret);
	if (err) return err;

	((Object *)*ret)->parent=owner;

	return ERR_OK;
}

Error Remove(PWObject object)
{
	return PWRelease(object,object);
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_KEYPRESS_ACTION:
			obj->Action=(Error (*)(PWObject))(*(taglist+1));
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}


Error Handle(PWObject object, PWWait event)
{
	Object *obj=(Object *)object;

	if (obj->Action)
		return obj->Action(object);
	else if (obj->parent && obj->parent->type->PWHandle)
		return obj->parent->type->PWHandle(obj->parent,PW_EVENT_HIT);

	return ERR_OK;
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, PW_TYPE_KEYPRESS, PW_TYPE_EVENT_KEYPRESS,
	&Create, &Change, NULL, &Remove, NULL, NULL, &Handle, NULL,
	NULL, NULL, NULL
	};
