/*
	ProWesS, PROGS Windowing System

	Label item object "LABL"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 February 1995
	15.10.1995 increased itemwidth by 2 pixels to prevent clipping
	22.03.1996 adjusted code for internal use of pixel values
	10.04.1996 use PWHandleDefine && use ProWesS defaults
*/

#include "mem_h"

#include "PWHandler_h"

#define TYPE PW_TYPE_LABEL

/* global statusses */

#define p100 short2pt(100)
ColourRGB ink = { 0, p100, 0 }; char Kink=FALSE;

char title_font[PF_MAXFONTNAME]=""; /* name of PROforma font to use */
pt title_fontsize=0;				/* fontsize to use */
pt title_baseline=0;				/* font baseline position (at fontsize) */
pt title_height=0;					/* height of font (at fontsize) */
char recalculate=TRUE;

/* loose item object definition */

typedef struct {
	PWBaseObject;
	char *item; 					/* item identifier */
	pt itemwidth;
} Object;

/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	LinkPROforma();
	/* allocate specific object - which should be cleared (automatic) */
	return PWAllocate(event.region->system,sizeof(Object),ret);
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_LABEL_TEXT :
			obj->item=(char *)(*(taglist+1));
			system->status|=SSTATUS_WINCHANGE;
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Error Remove(PWObject object)
{
	return PWRelease(object,object);
}

Error Draw(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *reg=obj->event.region;
	Gstate gstate=reg->system->gstate;
	pix hit_xsiz=reg->hit_xend-reg->hit_xorg;
	pix hit_ysiz=reg->hit_yend-reg->hit_yorg;

	if (!gstate) return ERR_OK;

	PFPWindowSubSet(gstate,hit_xsiz,hit_ysiz,reg->hit_xorg,reg->hit_yorg);
	PFColourRGB(gstate,&ink);
	PFFontLoad(gstate,title_font);
	PFFontScale(gstate,title_fontsize);
	if (reg->level&1)
		/* left of item */
		PFMoveTo(gstate,xpix2PF(hit_xsiz)-obj->itemwidth-xpix2PF(1),title_baseline);
	else
		/* above item */
		PFMoveTo(gstate,xpix2PF(1),ypix2PF(hit_ysiz)-title_height+title_baseline);
	PFStringShow(gstate,obj->item);
	PFWindowSubRestore(gstate);

	return ERR_OK;
}

Defines defines[]={
{ "INK-COLOUR", PW_DEFINE_COLOUR,   NULL,   &ink,           &Kink },
{ "FONT",       PW_DEFINE_FONT,     NULL,   title_font,     &recalculate },
{ "FONTSIZE",   PW_DEFINE_PT,       NULL,   &title_fontsize,&recalculate },
{ 0 }
};

Error Define(char *name, char *value)
{
	return PWHandleDefine("LABEL-",defines,name,value);
}

Error DetermineSize(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *region=obj->event.region;
	Gstate gstate=region->system->gstate;

	region->hit_off_lft=region->hit_off_rgt=
	region->hit_off_top=region->hit_off_bot=0;

	if (!gstate) return ERR_OK;

	if (!Kink) PWQuery(object,PW_DEFAULT_MIDDLEGROUND,&ink);
	if (recalculate)
	{
		Bbox bbox;
		recalculate=FALSE;
		if (!title_fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&title_fontsize);
		if (!*title_font) PWQuery(object,PW_DEFAULT_FONT,title_font);
		PFFontLoad(gstate,title_font);
		PFFontScale(gstate,title_fontsize);
		PFFontBbox(gstate,&bbox);
		title_baseline=bbox.yorg;
		title_height=bbox.ysiz;
	}

	/* set itemwidth */
	PFFontLoad(gstate,title_font);
	PFFontScale(gstate,title_fontsize);
	PFStringWidth(gstate,obj->item,&obj->itemwidth);

	region->tot_ysiz=short2pt(PF2ypix(title_height));
	region->tot_xsiz=short2pt(PF2xpix(obj->itemwidth)+2);

	return ERR_OK;
}


PWTypeDef init={
	PW_TYPEDEF_FLAG, TYPE, PW_TYPE_EVENT_REGION,
	&Create, &Change, NULL, &Remove, NULL, NULL, NULL, &Define,
	&Draw, &DetermineSize, NULL
};
