/*
	ProWesS, PROGS Windowing System

	loose item object with PROforma text "LOOS"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 January 1995
	14.10.1995 made sure itemwidth always 2 pixels too large
	28.10.1995 loose_text didn't redraw when not autosize - fixed
	01.11.1995 made sure string always drawn centered when it should be - also
			   when autosize==FALSE
	14.11.1995 fixed bug where infinite loop can be caused by a keypress
			   attached to an unavailable item
	18.12.1995 now also use Draw routine from PW_TYPE_iTEM
	21.03.1996 internal data now more pixel oriented
	10.04.1996 use PWHandleDefine && use ProWesS default font & fontsize
	11.06.1996 PW_LOOSE_AUTOREPEAT added (large part thanks to Wolfgang Lenerz)
*/

#include "mem_h"
#include "str_h"

#include "PWHandler_h"

typedef struct {
	ColourRGB paper;
	ColourRGB ink;
	} Colours;

/* global statusses */

#define p100 short2pt(100)
Colours col_available  ={ {p100,p100,p100},{0,0,0}};
Colours col_unavailable={ {p100,p100,p100},{0,short2pt(52),0}};
Colours col_selected   ={ {0,0,0		 },{p100,p100,p100}};

char loose_font[PF_MAXFONTNAME]=""; /* name of PROforma font to use */
pt loose_fontsize=0;				/* fontsize to use */
pt loose_baseline=0;				/* font baseline position (at fontsize) */
pt loose_height=0;					/* height of font (at fontsize) */
char recalculate=TRUE;
TimeOut dragtimeout=5;				/* timeout for repeat on keydown */

PWTypeDef *border;

/* loose item object definition */

typedef struct {
	PWBaseObject;
	PWObject keypress;				/* keypress associated with item */
	char istatus;					/* status for this object */
	char change;					/* TRUE if status should change on event */
	char windowdo;					/* activate DO keypress in window ? */
	char center;					/* center item in object */
	char autosize;					/* modify item size when text changes ?? */
	char copied;					/* TRUE when string allocated by type */
	char whatdrag;					/* flag HITDRAG or DODRAG ? */
	char *item; 					/* item identifier */
	pt itemwidth;					/* width of item */
	pt itemheight;					/* height of item */
	Error (*HitAction)(PWObject object);	/* action for HIT on object */
	Error (*DoAction)(PWObject object); 	/* action for DO  on object */
	Error (*DrawAction)(PWObject object, Gstate gstate,
						pt xsiz, pt ysiz);	/* action for drawing of object */
	} Object;

/* object handling routines */


Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Error err;

	LinkPROforma();

	err=PWGetType(PW_TYPE_iTEM, &border);
	if (err) return err;

	/* allocate specific object - which should be cleared (automatic) */
	err=PWAllocate(event.region->system,sizeof(Object),ret);
	if (err) return err;

	/* make sure the item is available */
	(*(Object **)ret)->istatus=(char)PW_STATUS_AVAILABLE;
	(*(Object **)ret)->change=TRUE;
	(*(Object **)ret)->center=TRUE;
	(*(Object **)ret)->autosize=TRUE;
	event.region->wait_event=PW_EVENT_HIT|PW_EVENT_DO;

	return err;
}

Error Remove(PWObject object)
{
	Object *obj=(Object *)object;
	Heap heap=object->event.region->system->heap;

	if (obj->keypress) PWRemove(obj->keypress);
	if (obj->copied) PWRelease((PWObject)obj,obj->item);
	return PWRelease((PWObject)obj,(PWObject)obj);
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	Gstate gstate=system->gstate;
	char unrecognised_tag=FALSE;

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_LOOSE_STATUS :
			if (obj->istatus!=(char)(*(taglist+1)))
			{
				obj->istatus=(char)(*(taglist+1));
				obj->status|=OSTATUS_REDRAW;
				system->status|=SSTATUS_REDRAW;
			}
			break;
		case PW_LOOSE_TEXT :
			if (obj->copied) PWRelease((PWObject)obj,obj->item);
			obj->item=(char *)(*(taglist+1));
autosize:	if (obj->autosize)
			{
				obj->itemwidth=0;
				system->status|=SSTATUS_WINCHANGE;
			}
			else
			{
				obj->status|=OSTATUS_REDRAW;
				system->status|=SSTATUS_REDRAW;
			}
			break;
		case PW_LOOSE_TEXT_COPY:
			if (obj->copied) PWRelease((PWObject)obj,obj->item);
			err=PWAllocate(system,strlen((char *)(*(taglist+1)))+1,&obj->item);
			if (err) { obj->item=NULL; return err; }
			STRCopy((char *)(*(taglist+1)),obj->item);
			goto autosize;
		case PW_LOOSE_CHANGE_STATUS:
			obj->change=(char)(*(taglist+1));
			break;
		case PW_LOOSE_WINDOW_DO:
			obj->windowdo=(char)(*(taglist+1));
			break;
		case PW_LOOSE_CENTER_ITEM:
			if (obj->center!=(char)(*(taglist+1)))
			{
				obj->center=(char)(*(taglist+1));
				obj->status|=OSTATUS_REDRAW;
				system->status|=SSTATUS_REDRAW;
			}
			break;
		case PW_LOOSE_AUTOSIZE:
			obj->autosize=(char)(*(taglist+1));
			if (obj->autosize)
			{
				obj->itemheight=0;
				obj->itemwidth=0;
				system->status|=SSTATUS_WINCHANGE;
			}
			break;
		case PW_LOOSE_ACTION_HIT:
			obj->HitAction=(Error (*)(PWObject))(*(taglist+1));
			break;
		case PW_LOOSE_ACTION_DO:
			obj->DoAction=(Error (*)(PWObject))(*(taglist+1));
			break;
		case PW_LOOSE_ACTION_DRAW:
			obj->DrawAction=(Error (*)(PWObject,Gstate,pt,pt))(*(taglist+1));
			break;
		case PW_LOOSE_SIZE:
			if (obj->itemwidth!=(*(taglist+1)) || obj->itemheight!=(*(taglist+2)))
			{
				obj->itemwidth=(*(taglist+1));
				obj->itemheight=(*(taglist+2));
				system->status|=SSTATUS_WINCHANGE;
				obj->autosize=FALSE;
			}
			break;
		case PW_LOOSE_XSIZE:
			if (obj->itemwidth!=(*(taglist+1)))
			{
				obj->itemwidth=(*(taglist+1));
				system->status|=SSTATUS_WINCHANGE;
				obj->autosize=FALSE;
			}
			break;
		case PW_LOOSE_YSIZE:
			if (obj->itemheight!=(*(taglist+1)))
			{
				obj->itemheight=(*(taglist+1));
				system->status|=SSTATUS_WINCHANGE;
				obj->autosize=FALSE;
			}
			break;
		case PW_LOOSE_KEYPRESS:
			/* attach keypress */
			if (!obj->keypress)
			{
				err=PWCreate(object,&obj->keypress,PW_TYPE_KEYPRESS,NULL);
				if (err) return err;
			}
			err=PWChange(obj->keypress,PW_KEYPRESS,(*(taglist+1)),NULL);
			if (err) return err;
			break;
		case PW_LOOSE_AUTOREPEAT:
			obj->event.region->wait_event|=PW_EVENT_HITDRAG|PW_EVENT_DODRAG|PW_EVENT_TIMEOUT;
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Error Query(PWObject object, int what, void *ret)
{
	Error err=ERR_OK;
	Object *obj=(Object *)object;

	switch (what)
	{
		case PW_LOOSE_STATUS:
			*((int *)ret)=obj->istatus;
			break;
		case PW_LOOSE_TEXT:
			*((char **)ret)=obj->item;
			break;
		default :
			err=ERR_IPAR;
	}
	return err;
}

Error Draw(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *reg=obj->event.region;
	Gstate gstate=reg->system->gstate;
	pix hit_xsiz=reg->hit_xend-reg->hit_xorg;
	pix hit_ysiz=reg->hit_yend-reg->hit_yorg;
	Colours *col;
	Error err=ERR_OK;

	if (obj->istatus==PW_STATUS_AVAILABLE)
		col=&col_available;
	else if (obj->istatus==PW_STATUS_UNAVAILABLE)
		col=&col_unavailable;
	else
		col=&col_selected;

	if (!gstate) return err;

	PFPWindowSubSet(gstate,hit_xsiz,hit_ysiz,reg->hit_xorg,reg->hit_yorg);
	PFPaperColourRGB(gstate,&col->paper);
	PFColourRGB(gstate,&col->ink);
	PFPageClear(gstate);

	if (obj->item)
	{
		PFFontLoad(gstate,loose_font);
		PFFontScale(gstate,loose_fontsize);

		if (obj->center)
		{
			pt strwidth;
			PFStringWidth(gstate,obj->item,&strwidth);
			PFMoveTo(gstate,(xpix2PF(hit_xsiz)-strwidth)/2,
							(ypix2PF(hit_ysiz)-loose_height)/2+loose_baseline);
		}
		else
			PFMoveTo(gstate,xpix2PF(1),loose_baseline);
		
		PFStringShow(gstate,obj->item);
	}

	if (obj->DrawAction)
		err=obj->DrawAction((PWObject)obj, gstate,
							xpix2PF(hit_xsiz), xpix2PF(hit_ysiz));

	PFWindowSubRestore(gstate);
	border->PWDraw((PWObject)obj);

	return err;
}

Error Entry(PWObject object)
{
	Object *obj=(Object *)object;

	if (obj->istatus!=PW_STATUS_UNAVAILABLE)
		border->PWRegionEntry(object);

	return ERR_OK;
}

Error Exit(PWObject object)
{
	return border->PWRegionExit(object);
}

Error Handle(PWObject object, PWWait event)
{
	PWSystem *system=object->event.region->system;
	Object *obj=(Object *)object;
	Error err=ERR_OK;

	switch(event)
	{
	case PW_EVENT_DO:
		if (obj->istatus==(char)PW_STATUS_UNAVAILABLE)
			return PWHandleKeyPress(system);
		if (obj->change)
		{
			obj->istatus=PW_STATUS_SELECTED;
			Draw(object);
			PFPageShow(system->gstate);
			system->status&=~SSTATUS_PAGESHOW;
		}
		if (obj->DoAction)
			err=obj->DoAction(object);
		else if (obj->HitAction)
			err=obj->HitAction(object);
		if (err) return err;
		if (obj->windowdo)
			err=PWHandleKeyPress(system);
		break;
	case PW_EVENT_HIT:
		if (obj->istatus==(char)PW_STATUS_UNAVAILABLE)
			return system->ptr_info.keystroke==KEYSTROKE_HIT ?
				   PWHandleKeyPress(system) : ERR_OK;
		if (obj->change)
		{
			obj->istatus^=PW_STATUS_SELECTED;
			Draw(object);
			PFPageShow(system->gstate);
			system->status&=~SSTATUS_PAGESHOW;
		}
		if (obj->HitAction)
			err=obj->HitAction(object);
		break;
	case PW_EVENT_HITDRAG:
		obj->whatdrag=PW_EVENT_HIT;
		obj->event.region->timeout=dragtimeout;
		Handle((PWObject)obj,obj->whatdrag);
		break;
	case PW_EVENT_DODRAG:
		obj->whatdrag=PW_EVENT_DO;
		obj->event.region->timeout=dragtimeout;
		Handle((PWObject)obj,obj->whatdrag);
		break;
	case PW_EVENT_TIMEOUT:
		Handle((PWObject)obj,obj->whatdrag);
		break;
	case PW_EVENT_DRAGEND:
		obj->event.region->timeout=FOREVER;
		break;
	}
	return err;
}

Defines defines[]={
{ "AVAILABLE-INK-COLOUR",       PW_DEFINE_COLOUR,   NULL,   &col_available.ink,     NULL },
{ "UNAVAILABLE-INK-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &col_unavailable.ink,   NULL },
{ "SELECTED-INK-COLOUR",        PW_DEFINE_COLOUR,   NULL,   &col_selected.ink,      NULL },
{ "AVAILABLE-PAPER-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &col_available.paper,   NULL },
{ "UNAVAILABLE-PAPER-COLOUR",   PW_DEFINE_COLOUR,   NULL,   &col_unavailable.paper, NULL },
{ "SELECTED-PAPER-COLOUR",      PW_DEFINE_COLOUR,   NULL,   &col_selected.paper,    NULL },
{ "FONT",                       PW_DEFINE_FONT,     NULL,   loose_font,             &recalculate },
{ "FONTSIZE",                   PW_DEFINE_PT,       NULL,   &loose_fontsize,        &recalculate },
{ "AUTOREPEAT-TIMEOUT",         PW_DEFINE_SHORT,    NULL,   &dragtimeout,           NULL },
{ 0 }
};

Error Define(char *name, char *value)
{
	return PWHandleDefine("LOOSE-ITEM-",defines,name,value);
}

Error DetermineSize(PWObject object)
{
	PWRegion *region=object->event.region;
	Object *obj=(Object *)object;
	Gstate gstate=obj->event.region->system->gstate;

	if (!gstate) return ERR_OK;

	if (recalculate)
	{
		Bbox bbox;
		recalculate=FALSE;
		if (!loose_fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&loose_fontsize);
		if (!*loose_font) PWQuery(object,PW_DEFAULT_FONT,loose_font);
		PFFontLoad(gstate,loose_font);
		PFFontScale(gstate,loose_fontsize);
		PFFontBbox(gstate,&bbox);
		loose_baseline=bbox.yorg;
		loose_height=bbox.ysiz;
	}
	if (obj->itemwidth==0 || obj->autosize)
	{
		PFFontLoad(gstate,loose_font);
		PFFontScale(gstate,loose_fontsize);
		PFStringWidth(gstate,obj->item,&obj->itemwidth);
	}
	if (obj->itemheight==0 || obj->autosize)
		obj->itemheight=loose_height;
	region->tot_ysiz=short2pt(PF2ypix(obj->itemheight))+pt_half;
	region->tot_xsiz=short2pt(PF2xpix(obj->itemwidth)+2);

	return border->PWDetermineSize((PWObject)obj);
}


PWTypeDef init={
	PW_TYPEDEF_FLAG, PW_TYPE_LOOSE_ITEM, PW_TYPE_EVENT_REGION,
	&Create, &Change, &Query, &Remove, &Entry, &Exit, &Handle, &Define,
	&Draw, &DetermineSize, NULL
};
