/*
	ProWesS, PROGS Windowing System

	menu type for ProWesS, "MENU"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 16 July 1995
	14.10.1995 made sure string have a pixel at either side in menu
	23.03.1996 adjusted for internal use of pixel values
	10.04.1996 use PWHandleDefine && use ProWesS default font & fontsize
	02.06.1996 PW_MENU_ADD and variants now set the current item to allow the
			   user to change the status using PW_MENU_STATUS_CURRENT
	06.06.1996 - can now use default colours when none given explicitly
			   - added PW_MENU_KEYPRESSES create tag and use provide standard
	11.06.1996 (thanks to Wolfgang Lenerz) added PW_MENU_FIRST_NUMBER,
			   PW_MENU_NEXT_NUMBER, PW_MENU_AVAILABLE_FIRST_NUMBER,
			   PW_MENU_AVAILABLE_NEXT_NUMBER, PW_MENU_SELECTED_FIRST_NUMBER,
			   PW_MENU_SELECTED_NEXT_NUMBER, PW_MENU_STATUS_ALL
	01.07.1996 PW_MENU_STATUS was bugged when PW_MENU_UNIQUE
*/

#include "mem_h"
#include "str_h"

#include "PWHandler_h"

#define TYPE PW_TYPE_MENU
#define catch(X) if (err=(X)) return err; else

typedef struct { ColourRGB ink, paper; } ColourStatus;

#define BLOCKINGFACTOR 50			/* no of Item objects allocated together */

/* global statusses */

#define p100 short2pt(100)
ColourRGB ink	= { 0,    0,    0    }; char Kink=   FALSE;
ColourRGB paper = { p100, p100, p100 }; char Kpaper= FALSE;
ColourRGB border= { p100, 0,    0    }; char Kborder=FALSE;
ColourStatus selected	={{ p100, p100, p100 },{ 0   , 0   , 0    }};
ColourStatus available	={{ 0	, 0   , 0	 },{ p100, p100, p100 }};
ColourStatus unavailable={{ 0	, p100, 0	 },{ p100, p100, p100 }};
char font[PF_MAXFONTNAME]="";		/* name of PROforma font to use */
pt fontsize=0;						/* fontsize to use */
pt baseline=0;						/* font baseline position (at fontsize) */
pix height=0, itemheight;			/* height of font (at fontsize) */
pix xwidth=0, ywidth=0; pt width;	/* width nominal and in each direction */
char recalculate=TRUE;

/* loose item object definition */

typedef struct _item {
	struct _item *next; 	/* next and .. */
	struct _item *prev; 	/* .. previous item in menu */
	char *item; 			/* string for this item */
	char status;			/* selected, available or unavailable ? */
	char copied;			/* was the item copied ? */
	} Item;

#define STATUS_MAXWIDTH 	(short)0x0001
#define STATUS_UNIQUE		(short)0x0002
#define STATUS_NOSELECT 	(short)0x0004
#define STATUS_WINDOWDO 	(short)0x0008

typedef struct {
	PWBaseObject;
	PWObject applic;		/* applic object connected with this menu */
	PWTypeDef *atype;		/* type definition for applic */
	Item *base, *sentinel;	/* base (before) and sentinel item */
	int count;				/* number of items in menu */
	int inrow;				/* no of items displayed side by side */
	Error (*Select)(PWObject,char *);	/* select action routine */
	Error (*DeSelect)(PWObject,char *); /* deselect action routine */
	pix itemwidth;			/* width for items */
	short mstatus;			/* several boolean statusses */
	Item *selected; 		/* currently selected object when STATUS_UNIQUE */
	int current;			/* count of item currently pointed */
	Error (*sortcmp)(char *s1, char *s2, int *res); /* cmp rout for sorting */
	pix canvaswidth;		/* current width of canvas */
	Item *qcurrent; 		/* current item for query operators */
	char ksu, ksd, klu, kld;/* keys for scrollbar */
	} Object;

/* forward declaration */
Error Remove(PWObject object);
Error add_string(Object *obj, char *str, char copy);

static void Calculate(Gstate gstate, PWObject object)
{
	Bbox bbox;

	if (!Kpaper) PWQuery(object,PW_DEFAULT_BACKGROUND,&paper);
	if (!Kink) PWQuery(object,PW_DEFAULT_FOREGROUND,&ink);
	if (!Kborder) PWQuery(object,PW_DEFAULT_MIDDLEGROUND,&border);
	if (!fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&fontsize);
	if (!*font) PWQuery(object,PW_DEFAULT_FONT,font);
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	PFFontBbox(gstate,&bbox);
	baseline=bbox.yorg;
	height=PF2ypix(bbox.ysiz);
	itemheight=height+ywidth;
	recalculate=FALSE;
}

/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Error err;
	Object *obj;
	int null=NUL;
	PWSystem *system=event.region->system;

	LinkPROforma();

	/* allocate specific object - which should be cleared (automatic) */
	catch( PWAllocate(event.region->system,sizeof(Object),ret) );

	obj=(Object *)*ret;
	if (recalculate) Calculate(event.region->system->gstate, (PWObject)obj);

	/* initialise object */
	err=PWGetType(PW_TYPE_APPLIC, &obj->atype);
	if (err) { Remove((PWObject)obj); return err; }

	err=obj->atype->PWCreate((PWObject)obj,&obj->applic,event,&null);
	if (err) { Remove((PWObject)obj); return err; }
	/* fill in the base object (normally done by system, which is bypassed) */
	obj->applic->type=obj->atype;
	obj->applic->event=event;

	obj->mstatus=STATUS_MAXWIDTH;	/* use maximum width of items */
	obj->inrow=1;					/* there is one item in each row */
	obj->current=-1;				/* no current item yet */
	/* set standard keypresses for scrolling the menu */
	obj->ksu=209; obj->ksd=217;
	obj->klu=213; obj->kld=221;

	/* initialise list of items (base and sentinel) */
	err=PWAllocate(system,sizeof(Item),&obj->base);
	if (err) { Remove((PWObject)obj); return err; }
	err=PWAllocate(system,sizeof(Item),&obj->sentinel);
	if (err) { Remove((PWObject)obj); return err; }
	obj->base->next=obj->sentinel;
	obj->sentinel->prev=obj->base;

	/* see if there is a PW_MENU_KEYPRESSES tag */
	while (*taglist)
	{
		if (*taglist==PW_MENU_KEYPRESSES)
		{
			/* set different keypresses for scrolling the menu */
			obj->ksu=(*(taglist+1)); obj->ksd=(*(taglist+2));
			obj->klu=(*(taglist+3)); obj->kld=(*(taglist+4));
			*taglist&=PW_TAGMASK_SKIP;
		}
		taglist=PWSkipTag(taglist);
	}


	return err;
}

Error Remove(PWObject object)
{
	Object *obj=(Object *)object;
	Item *list=obj->base->next;

	/* remove applic */
	if (obj->applic) obj->atype->PWRemove(obj->applic);

	/* release all allocates strings */
	if (obj->base && obj->sentinel)
		for (list=obj->base->next ; list->next ; list=list->next)
			if (list->copied)
				PWRelease((PWObject)obj,list->item);

	return PWRelease((PWObject)obj,(PWObject)obj);
}

/* event handler routines routines for canvas inside menu */

static Error DrawBorder(Object *obj, CanvasInfo *rinfo, ColourRGB *colour)
{
	PWRegion *reg=obj->event.region;
	Gstate gstate=rinfo->gstate;
	pix itemwidth=(rinfo->xpixsiz-xwidth)/obj->inrow-xwidth;
	int row=obj->current/obj->inrow-rinfo->yorg;
	int column=obj->current-(row+rinfo->yorg)*obj->inrow;
	pt xsiz=xpix2PF(itemwidth+xwidth);
	pt ysiz=ypix2PF(itemheight);
	pt xorg=xpix2PF((itemwidth+xwidth)*column*2+xwidth)/2;
	pt yorg=ypix2PF( itemheight 	  *row	 *2+ywidth)/2;

	if (row<0 || column<0) return ERR_OK;

	if (recalculate) Calculate(gstate,(PWObject)obj);

	PFPathStroked(gstate);
	PFColourRGB(gstate, colour);

	PFLineWidth(gstate, xpix2PF(xwidth));
	PFMoveTo(gstate, xorg	  , yorg);
	PFLineR(gstate, 0, ysiz);
	PFMoveTo(gstate, xorg+xsiz, yorg);
	PFLineR(gstate, 0, ysiz);
	PFPathDraw(gstate);

	PFLineWidth(gstate, ypix2PF(ywidth));
	PFMoveTo(gstate, xorg, yorg);
	PFLineR(gstate, xsiz, 0);
	PFMoveTo(gstate, xorg, yorg+ysiz);
	PFLineR(gstate, xsiz, 0);
	PFPathDraw(gstate);

	reg->system->status|=SSTATUS_PAGESHOW;

	return ERR_OK;
}
Error canvas_move(PWObject object, CanvasInfo *rinfo)
{
	Object *obj;
	int current=-1; 				/* current item - none just yet */
	int row,column;
	pix itemwidth;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);

	itemwidth=(rinfo->xpixsiz-xwidth)/obj->inrow-xwidth;

	/* calculate current item */
	column=rinfo->xpixpos/itemwidth;
	if (column>=obj->inrow) column=obj->inrow-1;
	row   =rinfo->ypixpos/itemheight;
	if (column*itemwidth +xwidth<=rinfo->xpixpos &&
		row   *itemheight+ywidth<=rinfo->ypixpos)
	{
		current=(rinfo->yorg+row)*obj->inrow+column;
		if (current>=obj->count) current=-1;
	}

	/* redraw border if current has changed */
	if (current!=obj->current)
	{
		if (obj->current!=-1)
			DrawBorder(obj,rinfo,&paper);
		obj->current=current;
		if (current!=-1)
			DrawBorder(obj,rinfo,&border);
	}
	return ERR_OK;
}
Error canvas_exit(PWObject object, CanvasInfo *rinfo)
{
	Object *obj;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);

	if (obj->current!=-1)
		DrawBorder(obj,rinfo,&paper);
	obj->current=-1;

	return ERR_OK;
}
Error canvas_scale(PWObject object, CanvasInfo *rinfo)
{
	Object *obj;
	int inrow;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);

	if (recalculate) Calculate(obj->event.region->system->gstate,object);

	obj->canvaswidth=rinfo->xpixsiz-2*xwidth;
	/* determine number of visible items and itemwidth */
	inrow=(rinfo->xpixsiz-xwidth)/(obj->itemwidth+xwidth);
	if (inrow<1 || obj->count==0) inrow=1;

	if (inrow!=obj->inrow)
	{
		obj->inrow=inrow;
		PWChange(obj->applic,
					PW_APPLIC_YSCROLL_LIST,
						PW_SCROLL_MAXIMUM, (obj->count+inrow-1)/inrow,
						PW_SCROLL_SCROLL, 0,
						NULL,
					NULL);
	}
#ifdef DEBUG
	/* when debugging, add item in menu which gives the size of canvas */
	{
		char string[30];
		STRFormatPutSafe(string,30,"size %d %d",rinfo->xpixsiz,rinfo->ypixsiz);
		add_string(obj,string,TRUE);
	}
#endif DEBUG
	return ERR_OK;
}
Error canvas_redraw(PWObject object, CanvasInfo *rinfo)
{
	Object *obj;
	int count, row, col;
	pix itemwidth;
	Item *item;
	ColourStatus *status;
	Gstate gstate=rinfo->gstate;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);

	if (recalculate) Calculate(gstate,object);

	/* determine number of visible items and itemwidth */
	itemwidth=(rinfo->xpixsiz-xwidth)/obj->inrow-xwidth;

	/* search for first item to display */
	for (item=obj->base->next,count=rinfo->yorg*obj->inrow ;
		 item->next && count ;
		 item=item->next,count--);

	/* clear drawing area and set some drawing parameters */
	PFPaperColourRGB(gstate,&paper);
	PFPageClear(gstate);
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);

	/* display all items until the menu is filled */
	for (row=col=0 ; item->next ; item=item->next)
	{
		/* determine colours to use for drawing */
		status=&selected;
		if (item->status==PW_STATUS_UNAVAILABLE) status=&unavailable;
		if (item->status==PW_STATUS_AVAILABLE) status=&available;
		/* draw item */
		PFPWindowSubSet(gstate,itemwidth,height,
							  col*(itemwidth+xwidth)+xwidth,
							  row* itemheight		+ywidth);
		PFPaperColourRGB(gstate,&status->paper);
		PFColourRGB(gstate,&status->ink);
		PFPageClear(gstate);
		PFMoveTo(gstate,xpix2PF(1),baseline);
		PFStringShow(gstate,item->item);
		PFWindowSubRestore(gstate);
		/* move to next position */
		col++;
		if (col==obj->inrow)
		{
			row++; col=0;
			if (row*itemheight+ywidth>rinfo->ypixsiz) break;
		}
	}

	/* allow redraw of border */
	obj->current=-1;
	return ERR_OK;
}
Error canvas_key(PWObject object, CanvasInfo *rinfo, PWWait event)
{
	Object *obj;
	Item *list;
	int i;
	Error err=ERR_OK;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);

	if (obj->current<0) goto exit_key;

	for (i=obj->current,list=obj->base->next ; i ; i--, list=list->next);
	if (!(obj->mstatus&STATUS_NOSELECT))
	{
		if (event==PW_EVENT_HIT)
		{
			if (list->status==PW_STATUS_UNAVAILABLE) goto exit_key;
			list->status^=PW_STATUS_SELECTED;
			if (list->status==PW_STATUS_AVAILABLE && obj->DeSelect)
				err=obj->DeSelect((PWObject)obj,list->item);
		}
		else
		{
			if (list->status!=PW_STATUS_AVAILABLE) goto exit_key;
			list->status=PW_STATUS_SELECTED;
		}
		if (list->status==PW_STATUS_SELECTED && obj->mstatus&STATUS_UNIQUE)
		{
			if (obj->selected && obj->selected!=list)
			{
				obj->selected->status=PW_STATUS_AVAILABLE;
				if (obj->DeSelect)
					err=obj->DeSelect((PWObject)obj,obj->selected->item);
			}
			obj->selected=list;
		}
		canvas_redraw(object,rinfo);
		canvas_move(object,rinfo);
	}
	if (list->status==PW_STATUS_SELECTED || obj->mstatus&STATUS_NOSELECT)
		if (obj->Select)
			err=obj->Select((PWObject)obj,list->item);
exit_key:
	if (event==PW_EVENT_DO && obj->mstatus&STATUS_WINDOWDO)
		PWHandleKeyPress(obj->event.region->system);
	return err;
}
Error canvas_hit(PWObject object, CanvasInfo *rinfo)
{
	return canvas_key(object,rinfo,PW_EVENT_HIT);
}
Error canvas_do(PWObject object, CanvasInfo *rinfo)
{
	return canvas_key(object,rinfo,PW_EVENT_DO);
}

/* event handler routines for scrollbar inside menu */

Error calcsize(PWObject object, pt *size)
{
	if (recalculate) Calculate(object->event.region->system->gstate,object);
	*size=(PF2ypix(*size)+1-ywidth)/itemheight;
	return ERR_OK;
}

/* routine too change the itemwidth */
Error change_width(Object *obj, pt _width)
{
	int inrow;
	pix width=PF2xpix(_width)+2;

	if (width>obj->itemwidth)
	{
		obj->itemwidth=width;
		if (width>obj->canvaswidth)
		{
			PWChange(obj->applic,
						PW_APPLIC_CANVAS_LIST,
							PW_CANVAS_XSIZE_PIX, width+xwidth*2,
							NULL,
						NULL);
		}
		if (recalculate) Calculate(obj->event.region->system->gstate,(PWObject)obj);
		/* determine number of visible items and itemwidth */
		inrow=(obj->canvaswidth+xwidth)/(obj->itemwidth+xwidth);
		if (inrow<1) inrow=1;

		if (inrow!=obj->inrow)
		{
			obj->inrow=inrow;
			PWChange(obj->applic,
						PW_APPLIC_YSCROLL_LIST,
							PW_SCROLL_MAXIMUM, (obj->count+inrow-1)/inrow,
							PW_SCROLL_SCROLL, 0,
							NULL,
						NULL);
		}
	}
	return ERR_OK;
}

/* routine to add a string in the object */

Error sort_add(Object *obj, Item *item)
{
	Item *curr=obj->base->next;
	int res;

	while(curr->next)
	{
		obj->sortcmp(curr->item,item->item,&res);
		if (res>0)
			break;
		else
			curr=curr->next;
	}
	item->prev=curr->prev;
	item->next=curr;
	curr->prev->next=item;
	curr->prev=item;
	return ERR_OK;
}

Error add_string(Object *obj, char *str, char copy)
{
	PWSystem *system=obj->event.region->system;
	Gstate gstate=system->gstate;
	Item *new;
	Error err;

	/* allocate and fill in */
	catch( PWAllocate(system,sizeof(Item),&new) );
	new->status=PW_STATUS_AVAILABLE;
	new->copied=copy;
	if (copy)
	{
		err=PWAllocate(system,strlen(str)+1,&new->item);
		if (err) { PWRelease((PWObject)obj,new); return err; }
		STRCopy(str,new->item);
	}
	else
		new->item=str;

	/* link in */
	if (!obj->sortcmp)
	{
		new->next=obj->sentinel;
		new->prev=obj->sentinel->prev;
		obj->sentinel->prev->next=new;
		obj->sentinel->prev=new;
	}
	else
	{
		sort_add(obj,new);
	}
	obj->qcurrent=new;				/* newly added item should be current */

	/* increase number of items in applic */
	PWChange(obj->applic,
					PW_APPLIC_YSCROLL_LIST,
						PW_SCROLL_MAXIMUM, ((obj->count++)+obj->inrow)/obj->inrow,
						NULL,
					PW_APPLIC_CANVAS_LIST,
						PW_CANVAS_REDRAW,
						NULL,
					NULL);

	/* set new width for items */
	if (obj->mstatus&STATUS_MAXWIDTH)
	{
		pt width;
		PFFontLoad(gstate,font);
		PFFontScale(gstate,fontsize);
		PFStringWidth(gstate,str,&width);
		change_width(obj,width);
	}

	return err;
}


Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;
	Gstate gstate=system->gstate;
	Heap heap=system->heap;

	if (recalculate) Calculate(system->gstate,(PWObject)obj);

	if (!(obj->status&OSTATUS_INITIALISED))
	{
		obj->status|=OSTATUS_INITIALISED;
		PWChange(obj->applic,
					PW_APPLIC_CANVAS_LIST,
						PW_OBJECT_AUXILIARY, obj,
						PW_CANVAS_YSIZE_PIX, itemheight+ywidth,
						PW_CANVAS_ACTION_REDRAW, &canvas_redraw,
						PW_CANVAS_ACTION_HIT, &canvas_hit,
						PW_CANVAS_ACTION_DO, &canvas_do,
						PW_CANVAS_ACTION_MOVE, &canvas_move,
						PW_CANVAS_ACTION_EXIT, &canvas_exit,
						PW_CANVAS_ACTION_SCALE, &canvas_scale,
						NULL,
					PW_APPLIC_SCROLLBAR_Y,
					PW_APPLIC_YSCROLL_LIST,
						PW_OBJECT_AUXILIARY, obj,
						PW_SCROLL_MINDIST, 1,
						PW_SCROLL_MAXDIST, -1,
						PW_SCROLL_CALCSIZE, &calcsize,
						PW_SCROLL_MIN_KEYPRESSES, obj->ksu, obj->ksd,
						PW_SCROLL_MAX_KEYPRESSES, obj->klu, obj->kld,
						NULL,
					NULL);
	}

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_MENU_ADD:
			catch( add_string(obj,(char *)(*(taglist+1)),FALSE) );
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_MENU_ADD_COPY:
			catch( add_string(obj,(char *)(*(taglist+1)),TRUE) );
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_MENU_ADD_ARRAY:
			{
				int i,count=(*(taglist+1));
				int length=(*(taglist+2));
				char *str=(char *)(*(taglist+3));
				for (i=count ; i ; i--, str+=length)
					catch( add_string(obj,str,FALSE) );
			}
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_MENU_STATUS_CURRENT:
			if (obj->mstatus&STATUS_NOSELECT &&
				(*(taglist+1))==PW_STATUS_SELECTED)
				break;
			if (obj->qcurrent)
				obj->qcurrent->status=(*(taglist+1));
			if (obj->mstatus&STATUS_UNIQUE &&
				(*(taglist+1))==PW_STATUS_SELECTED &&
				obj->selected!=obj->qcurrent)
			{
				if (obj->DeSelect)
					obj->DeSelect((PWObject)obj,obj->selected->item);
				obj->selected=obj->qcurrent;
			}
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_MENU_STATUS:
			if (obj->mstatus&STATUS_NOSELECT &&
				(*(taglist+1))==PW_STATUS_SELECTED)
				break;
			{
				Item *item=obj->base->next;
				while (item->next && item->item!=(char *)(*(taglist+2)))
					item=item->next;
				if (item->item==(char *)(*(taglist+2)))
					item->status=(*(taglist+1));
				if (obj->mstatus&STATUS_UNIQUE &&
					(*(taglist+1))==PW_STATUS_SELECTED &&
					obj->selected && obj->selected!=item)
				{
					if (obj->DeSelect)
						obj->DeSelect((PWObject)obj,obj->selected->item);
					obj->selected=item;
				}
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			}
			break;
		case PW_MENU_STATUS_ALL:
			/* can't select all if not all selectable */
			if (obj->mstatus&(STATUS_UNIQUE|STATUS_NOSELECT))
				break;
			{
				Item *item=obj->base->next;
				while (item->next)
				{
					item->status=(*(taglist+1));
					item=item->next;
				}
			}
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_MENU_ITEMWIDTH:
			obj->itemwidth=PF2xpix(*(taglist+1));
			goto setitemwidth;
		case PW_MENU_ITEMWIDTH_PIX:
			obj->itemwidth=(*(taglist+1));
			goto setitemwidth;
		case PW_MENU_ITEMWIDTH_FS:
			obj->itemwidth=PF2xpix(fixmul((*(taglist+1)),fontsize));
setitemwidth:
			obj->mstatus&=~STATUS_MAXWIDTH;
			PWChange(obj->applic,
						PW_APPLIC_CANVAS_LIST,
							PW_CANVAS_XSIZE_PIX, obj->itemwidth+xwidth*2,
							NULL,
						NULL);
		case PW_MENU_ITEMWIDTH_MAX:
			obj->mstatus|=STATUS_MAXWIDTH;
			obj->itemwidth=0;
			{
				Item *list=obj->base->next;
				pt width;
				PFFontLoad(gstate,font);
				PFFontScale(gstate,fontsize);
				for (;list->next;list=list->next)
				{
					PFStringWidth(gstate,list->item,&width);
					change_width(obj,width);
				}
			}
			break;
		case PW_MENU_ACTION_SELECT:
			obj->Select=(Error (*)(PWObject,char *))(*(taglist+1));
			break;
		case PW_MENU_ACTION_DESELECT:
			obj->DeSelect=(Error (*)(PWObject,char *))(*(taglist+1));
			break;

		case PW_MENU_NONE_SELECTED:
			obj->mstatus|=STATUS_NOSELECT;
		case PW_MENU_UNIQUE:
			obj->mstatus|=STATUS_UNIQUE;
			obj->selected=NULL; 	/* no currently selected item */
			/* reset status of all objects to available */
			{
				Item *list=obj->base->next;
				for (;list->next;list=list->next)
					list->status=PW_STATUS_AVAILABLE;
			}
			PWChange(obj->applic,
						PW_APPLIC_CANVAS_LIST,
							PW_CANVAS_REDRAW,
							NULL,
						NULL);
			break;

		case PW_MENU_SORT_COMPARE:
			obj->sortcmp=(Error (*)(char *,char *,int *))(*(taglist+1));
			if (obj->sortcmp && obj->base->next->next)
			{
				/* resort the list with compare routine */
				Item *list=obj->base->next;

				obj->sentinel->prev->next=NULL;
				obj->base->next=obj->sentinel;
				obj->sentinel->prev=obj->base;

				while(list)
				{
					Item *curr=list;
					list=list->next;
					sort_add(obj,curr);
				}
			}
			break;
		case PW_MENU_VISIBLE_LINES:
			PWChange(obj->applic,
					PW_APPLIC_CANVAS_LIST,
						PW_CANVAS_YSIZE_PIX, (*(taglist+1))*itemheight+ywidth,
						NULL,
					NULL);
			break;
		case PW_MENU_CLEAR:
			/* release all allocated strings and the list */
			{
				Item *it, *base=obj->base;
				while (base->next->next)
				{
					it=base->next;
					if (it->copied)
						PWRelease((PWObject)obj,it->item);
					base->next=it->next;
					PWRelease((PWObject)obj,it);
				}
			}
			obj->sentinel->prev=obj->base;
			obj->count=0;
			obj->qcurrent=NULL;
			/* reset number of items in applic */
			PWChange(obj->applic,
							PW_APPLIC_YSCROLL_LIST,
								PW_SCROLL_MAXIMUM, 0,
								PW_SCROLL_SCROLL, 0,
								NULL,
							PW_APPLIC_CANVAS_LIST,
								PW_CANVAS_REDRAW,
								NULL,
							NULL);
			/* reset itemwidth ?? */
			if (obj->mstatus&STATUS_MAXWIDTH)
				obj->itemwidth=0;
			break;
		case PW_MENU_WINDOW_DO:
			if (*(taglist+1))
				obj->mstatus|=STATUS_WINDOWDO;
			else
				obj->mstatus&=~STATUS_WINDOWDO;
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

void cwidth(void *data)
{
	xwidth=PF2xpix(width); if (xwidth<1) xwidth=1;
	ywidth=PF2ypix(width); if (ywidth<1) ywidth=1;
}

Defines defines[]={
{ "BORDER-WIDTH",               PW_DEFINE_PT,       cwidth, &width,             NULL },
{ "AVAILABLE-INK-COLOUR",       PW_DEFINE_COLOUR,   NULL,   &available.ink,     NULL },
{ "UNAVAILABLE-INK-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &unavailable.ink,   NULL },
{ "SELECTED-INK-COLOUR",        PW_DEFINE_COLOUR,   NULL,   &selected.ink,      NULL },
{ "AVAILABLE-PAPER-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &available.paper,   NULL },
{ "UNAVAILABLE-PAPER-COLOUR",   PW_DEFINE_COLOUR,   NULL,   &unavailable.paper, NULL },
{ "SELECTED-PAPER-COLOUR",      PW_DEFINE_COLOUR,   NULL,   &selected.paper,    NULL },
{ "PAPER-COLOUR",               PW_DEFINE_COLOUR,   NULL,   &paper,             &Kpaper },
{ "INK-COLOUR",                 PW_DEFINE_COLOUR,   NULL,   &ink,               &Kink },
{ "BORDER-COLOUR",              PW_DEFINE_COLOUR,   NULL,   &border,            &Kborder },
{ "FONT",                       PW_DEFINE_FONT,     NULL,   font,               &recalculate },
{ "FONTSIZE",                   PW_DEFINE_PT,       NULL,   &fontsize,          &recalculate },
{ 0 }
};

Error Define(char *name, char *value)
{
	return PWHandleDefine("MENU-",defines,name,value);
}

Error Query(PWObject object, int what, void *_ret)
{
	Error err=ERR_OK;
	Object *obj=(Object *)object;
	char **ret=(char **)_ret;
	int counter;

	switch (what)
	{
	case PW_MENU_FIRST:
		obj->qcurrent=obj->base;
	case PW_MENU_NEXT:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		if (obj->qcurrent->next->next)
		{
			obj->qcurrent=obj->qcurrent->next;
			*ret=obj->qcurrent->item;
		}
		else
			*ret=NULL;
		break;
	case PW_MENU_AVAILABLE_FIRST:
		obj->qcurrent=obj->base;
	case PW_MENU_AVAILABLE_NEXT:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		*ret=NULL;
		while (obj->qcurrent->next->next)
		{
			obj->qcurrent=obj->qcurrent->next;
			if (obj->qcurrent->status==PW_STATUS_AVAILABLE)
			{
				*ret=obj->qcurrent->item;
				break;
			}
		}
		break;
	case PW_MENU_SELECTED_FIRST:
		obj->qcurrent=obj->base;
	case PW_MENU_SELECTED_NEXT:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		*ret=NULL;
		while (obj->qcurrent->next->next)
		{
			obj->qcurrent=obj->qcurrent->next;
			if (obj->qcurrent->status==PW_STATUS_SELECTED)
			{
				*ret=obj->qcurrent->item;
				break;
			}
		}
		break;
	case PW_MENU_FIRST_NUMBER:
		obj->qcurrent=obj->base;
	case PW_MENU_NEXT_NUMBER:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		if (obj->qcurrent->next->next)
		{
			Item *item=obj->base->next;
			counter=0;
			obj->qcurrent=obj->qcurrent->next;

			while (item->next && item->item!=obj->qcurrent->item)
			{
				item=item->next;
				counter++;
			}
			if (item->item==obj->qcurrent->item)
				*(int *)ret=counter;
			else
				*(int *)ret=-1; 	/* oops! */
		}
		else
			*ret=NULL;
		break;

/* THE FOLLOWING RETURN THE NUMBER OF THE SELECTED/AVAILABLE ITEMS */
/* with the following rules: */
/* - numbering starts at 0 (i.e. if the very first menu item is selected */
/*	this returns 0) */
/* - if the end of the elements is reached, this returns -1  */

	case PW_MENU_AVAILABLE_FIRST_NUMBER:
		obj->qcurrent=obj->base;
	case PW_MENU_AVAILABLE_NEXT_NUMBER:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		counter=1;
		if (!obj->qcurrent->next->next)
		{
			*(int *)ret=-1;
			break;
		}
		while (obj->qcurrent->next->next && counter)
		{
			obj->qcurrent=obj->qcurrent->next;
			if (obj->qcurrent->status==PW_STATUS_AVAILABLE)
				counter=0;
		}
		if (obj->qcurrent->status!=PW_STATUS_AVAILABLE)
		{
			*(int *)ret=-1;
			break;
		}
		{
			Item *item=obj->base->next;
			counter=0;
			while (item->next && item->item!=obj->qcurrent->item)
			{
				item=item->next;
				counter++;
			}
		}
		*(int *)ret=counter;
		break;

	case PW_MENU_SELECTED_FIRST_NUMBER:
		obj->qcurrent=obj->base;
	case PW_MENU_SELECTED_NEXT_NUMBER:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		counter=1;
		if (!obj->qcurrent->next->next)
		{
			*(int *)ret=-1;
			break;
		}
		while (obj->qcurrent->next->next && counter)
		{
			obj->qcurrent=obj->qcurrent->next;
			if (obj->qcurrent->status==PW_STATUS_SELECTED)
				counter=0;
		}
		if (obj->qcurrent->status!=PW_STATUS_SELECTED)
		{
			*(int *)ret=-1;
			break;
		}
		{
			Item *item=obj->base->next;
			counter=0;
			while (item->next && item->item!=obj->qcurrent->item)
			{
				item=item->next;
				counter++;
			}
		}
		*(int *)ret=counter;
		break;
	default:
		err=ERR_IPAR;
		break;
	}
	return err;
}


PWTypeDef init={
	PW_TYPEDEF_FLAG, TYPE, PW_TYPE_EVENT_REGION,
	&Create, &Change, &Query, &Remove, NULL, NULL, NULL, &Define,
	NULL, NULL, NULL
};
